# -*- coding: utf-8 -*-

import sys
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib
import matplotlib.textpath
import matplotlib.colors
import matplotlib.gridspec
import matplotlib.cm
import mpl_toolkits.axisartist as axisartist
from matplotlib.patches import Rectangle
import textwrap
import os.path


import scipy.sparse
from collections import OrderedDict

from intervaltree import IntervalTree, Interval

import hicexplorer.HiCMatrix as HiCMatrix
import hicexplorer.utilities

reload(sys)
sys.setdefaultencoding('utf-8')

DEFAULT_BED_COLOR = '#1f78b4'
DEFAULT_BIGWIG_COLOR = '#33a02c'
DEFAULT_BEDGRAPH_COLOR = '#a6cee3'
DEFAULT_MATRIX_COLORMAP = 'RdYlBu_r'
DEFAULT_TRACK_HEIGHT = 3  # in centimeters
DEFAULT_FIGURE_WIDTH = 40  # in centimeters
# proportion of width dedicated to (figure, legends)
DEFAULT_WIDTH_RATIOS = (0.93, 0.07)
DEFAULT_MARGINS = {'left': 0.04, 'right': 0.92, 'bottom': 0.12, 'top': 0.9}


class MultiDict(OrderedDict):
    """
    Class to allow identically named
    sections in configuration file
    by appending the section number as
    for example:
    1. section name
    """
    _unique = 0

    def __setitem__(self, key, val):
        if isinstance(val, OrderedDict):
            self._unique += 1
            key = "{}. [{}]".format(str(self._unique), key)
        OrderedDict.__setitem__(self, key, val)


class PlotTracks(object):

    def __init__(self, tracks_file, fig_width=DEFAULT_FIGURE_WIDTH,
                 fig_height=None, fontsize=None, dpi=None, track_label_width=None):
        self.fig_width = fig_width
        self.fig_height = fig_height
        self.dpi = dpi
        self.vlines_intval_tree = None
        self.vlines_properties = None
        start = self.print_elapsed(None)
        self.parse_tracks(tracks_file)
        if fontsize:
            fontsize = fontsize
        else:
            fontsize = float(fig_width) * 0.3
        if track_label_width is None:
            self.width_ratios = DEFAULT_WIDTH_RATIOS
        else:
            self.width_ratios = (1 - track_label_width, track_label_width)

        font = {'size': fontsize}
        matplotlib.rc('font', **font)
        # initialize each track
        self.track_obj_list = []
        for idx, properties in enumerate(self.track_list):
            if 'spacer' in properties:
                self.track_obj_list.append(PlotSpacer(properties))
                continue
            elif 'x-axis' in properties:
                self.track_obj_list.append(PlotXAxis(properties))
                continue
            if properties['file_type'] == 'bedgraph':
                self.track_obj_list.append(PlotBedGraph(properties))

            elif properties['file_type'] == 'bigwig':
                self.track_obj_list.append(PlotBigWig(properties))

            elif properties['file_type'] == 'bedgraph_matrix':
                self.track_obj_list.append(PlotBedGraphMatrix(properties))

            elif properties['file_type'] == 'hic_matrix':
                self.track_obj_list.append(PlotHiCMatrix(properties))

            elif properties['file_type'] == 'bed':
                self.track_obj_list.append(PlotBed(properties))

            elif properties['file_type'] == 'links':
                self.track_obj_list.append(PlotArcs(properties))

            elif properties['file_type'] == 'boundaries':
                self.track_obj_list.append(PlotBoundaries(properties))

            if 'title' in properties:
                # adjust titles that are too long
                # if the track label space is small
                if track_label_width < 0.1:
                    properties['title'] = textwrap.fill(properties['title'].encode("UTF-8"), 12)
                else:
                    properties['title'] = textwrap.fill(properties['title'].encode("UTF-8"), 50)

        print "time initializing track(s):"
        self.print_elapsed(start)

    def get_tracks_height(self, start, end):
        # prepare layout based on the tracks given.
        # The main purpose of the following loop is
        # to get the height of each of the tracks
        # because for the hi-C the height is variable with respect
        # to the range being plotted, the function is called
        # when each plot is going to be printed.
        track_height = []
        for track_dict in self.track_list:
            height = DEFAULT_TRACK_HEIGHT

            if 'width' in track_dict:
                height = track_dict['width']

            # compute the height of a Hi-C track
            # based on the depth such that the
            # resulting plot appears proportional
            #
            #      /|\
            #     / | \
            #    /  |d \   d is the depth that we want to be proportional
            #   /   |   \  when plotted in the figure
            # ------------------
            #   region len
            #
            # d (in cm) =  depth (in bp) * width (in cm) / region len (in bp)

            elif 'depth' in track_dict and track_dict['file_type'] == 'hic_matrix':
                # to compute the actual width of the figure the margins and the region
                # set for the legends have to be considered
                # DEFAULT_MARGINS[1] - DEFAULT_MARGINS[0] is the proportion of plotting area

                hic_width = \
                    self.fig_width * (DEFAULT_MARGINS['right'] - DEFAULT_MARGINS['left']) * self.width_ratios[0]
                scale_factor = 0.6  # the scale factor is to obtain a 'pleasing' result.
                depth = min(track_dict['depth'], (end - start))

                height = scale_factor * depth * hic_width / (end - start)
            track_height.append(height)

        return track_height

    def plot(self, file_name, chrom, start, end, title=None):
        track_height = self.get_tracks_height(start, end)

        if self.fig_height:
            fig_height = self.fig_height
        else:
            fig_height = sum(track_height)

        sys.stderr.write("Figure size in cm is {} x {}. Dpi is set to {}\n".format(self.fig_width,
                                                                                   fig_height, self.dpi))
        fig = plt.figure(figsize=self.cm2inch(self.fig_width, fig_height))
        if title:
            fig.suptitle(title)

        grids = matplotlib.gridspec.GridSpec(len(track_height), 2,
                                             height_ratios=track_height,
                                             width_ratios=self.width_ratios)
        axis_list = []
        for idx, track in enumerate(self.track_obj_list):
            axis = axisartist.Subplot(fig, grids[idx, 0])
            fig.add_subplot(axis)
            axis.axis[:].set_visible(False)
            # to make the background transparent
            axis.patch.set_visible(False)
            label_axis = plt.subplot(grids[idx, 1])
            label_axis.set_axis_off()
            track.plot(axis, label_axis, chrom, start, end)
            axis_list.append(axis)

        if self.vlines_intval_tree:
            self.plot_vlines(axis_list, chrom, start, end)

        fig.subplots_adjust(wspace=0, hspace=0.0,
                            left=DEFAULT_MARGINS['left'],
                            right=DEFAULT_MARGINS['right'],
                            bottom=DEFAULT_MARGINS['bottom'],
                            top=DEFAULT_MARGINS['top'])

        fig.savefig(file_name, dpi=self.dpi, transparent=False)
        return fig.get_size_inches()

    def plot_vlines(self, axis_list, chrom_region, start_region, end_region):
        """
        Plots dotted lines from the top of the first plot to the bottom
        of the last plot at the specified positions.

        :param axis_list: list of plotted axis
        :param chrom_region chromosome name
        :param start_region start position
        :param end_region end position

        :return: None
        """
        vlines_list = []
        if 'line width' in self.vlines_properties:
            line_width = self.vlines_properties['line width']
        else:
            line_width = 0.5

        if chrom_region not in self.vlines_intval_tree.keys():
            chrom_region = change_chrom_names(chrom_region)
        for region in sorted(self.vlines_intval_tree[chrom_region][start_region - 10000:end_region + 10000]):
            vlines_list.append(region.begin)

        for idx, track in enumerate(self.track_obj_list):
            ymin, ymax = axis_list[idx].get_ylim()

            axis_list[idx].vlines(vlines_list, ymin, ymax, linestyle='dashed', zorder=10,
                                  linewidth=line_width,
                                  color=(0, 0, 0, 0.7), alpha=0.5)

        return

    def parse_tracks(self, tracks_file):
        """
        Parses a configuration file

        :param tracks_file: file path containing the track configuration
        :return: array of dictionaries and vlines_file. One dictionary per track
        """
        from ConfigParser import SafeConfigParser
        from ast import literal_eval
        parser = SafeConfigParser(None, MultiDict)
        parser.readfp(open(tracks_file, 'r'))

        tracks_file_path = os.path.dirname(tracks_file)

        track_list = []
        for section_name in parser.sections():
            track_options = dict({"section_name": section_name})
            if section_name.endswith('[spacer]'):
                track_options['spacer'] = True
            elif section_name.endswith('[x-axis]'):
                track_options['x-axis'] = True
            for name, value in parser.items(section_name):
                if name in ['max_value', 'min_value', 'depth', 'width', 'line width', 'fontsize'] and value != 'auto':
                    track_options[name] = literal_eval(value)
                else:
                    track_options[name] = value

            if 'type' in track_options and track_options['type'] == 'vlines':
                self.vlines_properties = self.check_file_exists(track_options, tracks_file_path)
            elif 'skip' in track_options and track_options['skip'] != 'no':
                pass
            else:
                track_list.append(track_options)

        updated_track_list = []
        for track_dict in track_list:
            warn = None
            if 'file' in track_dict and track_dict['file'] != '':
                track_dict = self.check_file_exists(track_dict, tracks_file_path)
                if 'file_type' not in track_dict:
                    track_dict['file_type'] = self.guess_filetype(track_dict)

                #  set some default values
                if 'title' not in track_dict:
                    warn = "\ntitle not set for 'section {}'\n".format(track_dict['section_name'])
                    track_dict['title'] = ''
                if warn:
                    sys.stderr.write(warn)
            updated_track_list.append(track_dict)
        self.track_list = updated_track_list
        if self.vlines_properties:
            self.vlines_intval_tree, __, __ = file_to_intervaltree(self.vlines_properties['file'])

    @staticmethod
    def check_file_exists(track_dict, tracks_path):
        """
        Checks if a file or list of files exists. If the file does not exists
        tries to check if the file may be relative to the track_file path, in such case
        the path is updated.
        :param track_dict: dictionary of track values. Should contain
                            a 'file' key containing the path of the file
                            or files to be checked separated by space
                            For example: file1 file2 file3
        :param tracks_path: path of the tracks file
        :return: None
        """
        for file_type in ['boundaries_file', 'file']:
            if file_type in track_dict:
                file_names = [x for x in track_dict[file_type].split(" ") if x != '']
                full_path_file_names = []
                for file_name in file_names:
                    try:
                        open(file_name, 'r').close()
                        full_path_file_names.append(file_name)
                    except IOError:
                        try:
                            # try to find the file in the same path as the
                            # the path of the
                            name_with_tracks_path = tracks_path + "/" + file_name
                            open(name_with_tracks_path, 'r').close()
                            full_path_file_names.append(name_with_tracks_path)
                        except IOError:
                            sys.stderr.write("\n*ERROR*\nFile in section [{}] "
                                             "not found:\n{}\n\n".format(track_dict['section_name'],
                                                                         file_name))
                            sys.exit(1)

                track_dict[file_type] = " ".join(full_path_file_names)
        return track_dict

    @staticmethod
    def guess_filetype(track_dict):
        """

        :param track_dict: dictionary of track values with the 'file' key
                    containing a string path of the file or files. Only the ending
                     of the last file is used in case when there are more files
        :return: string file type detected
        """
        file = track_dict['file'].strip()
        if file.endswith(".bed") or file.endswith(".bed.gz"):
            file_type = 'bed'
        elif file.endswith(".npz") or file.endswith(".h5"):
            file_type = 'hic_matrix'
        elif file.endswith(".bw"):
            file_type = 'bigwig'
        elif file.endswith(".bg") or file.endswith(".bg.gz"):
            file_type = 'bedgraph'
        elif file.endswith(".bm") or file.endswith(".bm.gz"):
            file_type = 'bedgraph_matrix'
        else:
            sys.exit("Section {}: can not identify file type. Please specify "
                     "the file_type for {}".format(track_dict['section_name'], file))
        return file_type

    @staticmethod
    def cm2inch(*tupl):
        inch = 2.54
        if isinstance(tupl[0], tuple):
            return tuple(i / inch for i in tupl[0])
        else:
            return tuple(i / inch for i in tupl)

    @staticmethod
    def print_elapsed(start):
        import time
        if start:
            print time.time() - start
        return time.time()


def opener(filename):
    """
    Determines if a file is compressed or not
    """
    import gzip
    f = open(filename, 'rb')
    if f.read(2) == '\x1f\x8b':
        f.seek(0)
        return gzip.GzipFile(fileobj=f)
    else:
        f.seek(0)
        return f


def file_to_intervaltree(file_name):
    """
    converts a BED like file into a bx python interval tree
    :param file_name: string file name
    :return: interval tree dictionary. They key is the chromosome/contig name and the
    value is an IntervalTree. Each of the intervals have as 'value' the fields[3:] if any.
    """
    # iterate over a BED like file
    # saving the data into an interval tree
    # for quick retrieval
    file_h = opener(file_name)
    line_number = 0
    valid_intervals = 0
    prev_chrom = None
    prev_start = -1
    prev_line = None
    interval_tree = {}
    min_value = np.inf
    max_value = -np.inf

    for line in file_h.readlines():
        line_number += 1
        if line.startswith('browser') or line.startswith('track') or line.startswith('#'):
            continue
        fields = line.strip().split('\t')
        try:
            chrom, start, end = fields[0:3]
        except Exception as detail:
            msg = "Error reading line: {}\nError message: {}".format(line_number, detail)
            sys.exit(msg)

        try:
            start = int(start)
        except ValueError as detail:
            msg = "Error reading line: {}. The start field is not " \
                  "an integer.\nError message: {}".format(line_number, detail)
            sys.exit(msg)

        try:
            end = int(end)
        except ValueError as detail:
            msg = "Error reading line: {}. The end field is not " \
                  "an integer.\nError message: {}".format(line_number, detail)
            sys.exit(msg)

        if prev_chrom == chrom:
            assert prev_start <= start, \
                "Bed file not sorted. Please use a sorted bed file.\n{}{} ".format(prev_line, line)

        if chrom not in interval_tree:
            interval_tree[chrom] = IntervalTree()

        value = None

        if fields > 3:
            value = fields[3:]
            try:
                line_min = min(map(float, value))
                if line_min < min_value:
                    min_value = line_min

                line_max = max(map(float, value))
                if line_max > max_value:
                    max_value = line_max
            except ValueError:
                pass

        assert end > start, "Start position larger or equal than end for line\n{} ".format(line)

        interval_tree[chrom].add(Interval(start, end, value))
        valid_intervals += 1

    if valid_intervals == 0:
        sys.stderr.write("No valid intervals were found in file {}".format(file_name))
    file_h.close()

    return interval_tree, min_value, max_value


class TrackPlot(object):
    """
    The TrackPlot object is a holder for all tracks that are to be plotted.
    For example, to plot a bedgraph file a new class that extends TrackPlot
    should be created.

    It is expected that all TrackPlot objects have a plot method.

    """

    def __init__(self, properties_dict):
        self.properties = properties_dict


class PlotSpacer(TrackPlot):

    def plot(self, ax, label_ax, chrom_region, start_region, end_region):
        ax.set_xlim(start_region, end_region)
        pass


class PlotBedGraph(TrackPlot):

    def __init__(self, properties_dict):
        # super(self.__class__, self).__init__(*args, **kwargs)
        self.properties = properties_dict
        if 'color' not in self.properties:
            self.properties['color'] = DEFAULT_BEDGRAPH_COLOR
        self.interval_tree, ymin, ymax = file_to_intervaltree(self.properties['file'])

        if 'max_value' not in self.properties or self.properties['max_value'] == 'auto':
            self.properties['max_value'] = ymax

        if 'min_value' not in self.properties or self.properties['min_value'] == 'auto':
            self.properties['min_value'] = ymin

    def plot(self, ax, label_ax, chrom_region, start_region, end_region):
        self.ax = ax
        self.label_ax = label_ax
        score_list = []
        pos_list = []

        if chrom_region not in self.interval_tree.keys():
            chrom_region = change_chrom_names(chrom_region)

        for region in sorted(self.interval_tree[chrom_region][start_region - 10000:end_region + 10000]):
            score_list.append(float(region.data[0]))
            pos_list.append(region.begin + (region.end - region.begin) / 2)

        if 'color' not in self.properties:
            self.properties['color'] = DEFAULT_BEDGRAPH_COLOR

        if 'extra' in self.properties and self.properties['extra'][0] == '4C':
            # draw a vertical line for each fragment region center
            self.ax.fill_between(pos_list, score_list,
                                 facecolor=self.properties['color'],
                                 edgecolor='none')
            self.ax.vlines(pos_list, [0], score_list, color='olive', linewidth=0.5)
            self.ax.plot(pos_list, score_list, '-', color='slateblue', linewidth=0.7)
        else:
            try:
                self.ax.fill_between(pos_list, score_list, facecolor=self.properties['color'])
            except ValueError:
                sys.stderr.write("Invalid color {} for {}. "
                                 "Using gray instead.".format(self.properties['color'], self.properties['file']))
                self.ax.fill_between(pos_list, score_list, facecolor='gray')

        self.ax.set_frame_on(False)
        self.ax.axes.get_xaxis().set_visible(False)
        self.ax.axes.get_yaxis().set_visible(False)
        self.ax.set_xlim(start_region, end_region)

        ymax = self.properties['max_value']
        ymin = self.properties['min_value']

        if float(ymax) % 1 == 0:
            ymax_print = int(ymax)
        else:
            ymax_print = "{:.1f}".format(ymax)
        self.ax.set_ylim(ymin, ymax)
        ydelta = ymax - ymin
        small_x = 0.01 * (end_region - start_region)

        if 'show data range' in self.properties and \
                self.properties['show data range'] == 'no':
            pass
        else:
            # by default show the data range
            self.ax.text(start_region - small_x, ymax - ydelta * 0.2,
                         "[{}-{}]".format(ymin, ymax_print),
                         horizontalalignment='left', size='small',
                         verticalalignment='bottom')

        self.label_ax.text(0.15, 0.5, self.properties['title'],
                           horizontalalignment='left', size='large',
                           verticalalignment='center', transform=self.label_ax.transAxes)


class PlotBedGraphMatrix(PlotBedGraph):

    def plot(self, ax, label_ax, chrom_region, start_region, end_region):
        self.ax = ax
        self.label_ax = label_ax
        """
        Plots a bedgraph matrix file, that instead of having
        a single value per bin, it has several values.
        """

        start_pos = []
        matrix_rows = []
        if chrom_region not in self.interval_tree.keys():
            chrom_region = change_chrom_names(chrom_region)
        for region in sorted(self.interval_tree[chrom_region][start_region - 10000:end_region + 10000]):
            start_pos.append(region.begin)
            values = map(float, region.data)
            matrix_rows.append(values)

        matrix = np.vstack(matrix_rows).T
        if 'orientation' in self.properties and self.properties['orientation'] == 'inverted':
            matrix = np.flipud(matrix)

        if 'type' in self.properties and self.properties['type'] == 'lines':
            for row in matrix:
                self.ax.plot(start_pos, row, color='grey', linewidth=0.5)

            self.ax.plot(start_pos, matrix.mean(axis=0), linestyle="--", marker="|")
            ymax = self.properties['max_value']
            ymin = self.properties['min_value']
            self.ax.set_ylim(ymin, ymax)

            if 'show data range' in self.properties and self.properties['show data range'] == 'no':
                pass
            else:
                if float(ymax) % 1 == 0:
                    ymax_print = int(ymax)
                else:
                    ymax_print = "{:.1f}".format(ymax)

                if float(ymin) % 1 == 0:
                    ymin_print = int(ymin)
                else:
                    ymin_print = "{:.1f}".format(ymin)

                ydelta = ymax - ymin
                small_x = 0.01 * (end_region - start_region)
                # by default show the data range
                self.ax.text(start_region - small_x, ymax - ydelta * 0.2,
                             "[{}-{}]".format(ymin_print, ymax_print),
                             horizontalalignment='left',
                             verticalalignment='bottom')
            if 'plot horizontal lines' in self.properties and self.properties['horizontal lines']:
                # plot horizontal lines to compare values
                self.ax.hlines(np.arange(0, 1.1, 0.1), start_region, end_region, linestyle="--",
                               zorder=0, color='grey')

        else:
            x, y = np.meshgrid(start_pos, np.arange(matrix.shape[0]))
            shading = 'gouraud'
            vmax = self.properties['max_value']
            vmin = self.properties['min_value']

            img = self.ax.pcolormesh(x, y, matrix, vmin=vmin, vmax=vmax, shading=shading)
            img.set_rasterized(True)

        self.ax.set_xlim(start_region, end_region)
        self.ax.set_frame_on(False)
        self.ax.axes.get_xaxis().set_visible(False)
        self.ax.axes.get_yaxis().set_visible(False)
        self.properties['title'].replace('\\\\', '\\')
        self.label_ax.text(0.15, 0.5, self.properties['title'],
                           horizontalalignment='left', size='large',
                           verticalalignment='center', transform=self.label_ax.transAxes)


class PlotBigWig(TrackPlot):

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        import pyBigWig
        self.bw = pyBigWig.open(self.properties['file'])
        if 'color' not in self.properties:
            self.properties['color'] = DEFAULT_BIGWIG_COLOR

    def plot(self, ax, label_ax, chrom_region, start_region, end_region):
        self.ax = ax
        self.label_ax = label_ax
        formated_region = "{}:{}-{}".format(chrom_region, start_region, end_region)
        print "plotting {}".format(self.properties['file'])
        # compute the score in bins of 10000 SLOW
    #    score = np.array([self.bw.query(region[0], x, x+10000,1)[0]['mean']
    #                      for x in range(region[1], region[2], 10000)])

        num_bins = 700
        if 'number of bins' in self.properties:
            try:
                num_bins = int(self.properties['number of bins'])
            except TypeError:
                num_bins = 700
                sys.stderr.write("'number of bins' value: {} for bigwig file {} "
                                 "is not valid. Using default value (700)".format(self.properties['number of bins'],
                                                                                  self.properties['file']))

        if chrom_region not in self.bw.chroms().keys():
            chrom_region = change_chrom_names(chrom_region)

        if chrom_region not in self.bw.chroms().keys():
            sys.stderr.write("Can not read region {} from bigwig file:\n\n"
                             "{}\n\nPlease check that the chromosome name is part of the bigwig file "
                             "and that the region is valid".format(formated_region, self.properties['file']))

        if end_region - start_region < 2e6:
            scores = self.bw.values(chrom_region, start_region, end_region)

            if 'nans to zeros' in self.properties and self.properties['nans to zeros'] is True:
                scores[np.isnan(scores)] = 0

            scores = np.ma.masked_invalid(scores)

            lins = np.linspace(0, len(scores), num_bins).astype(int)
            scores_per_bin = [np.mean(scores[lins[x]:lins[x + 1]]) for x in range(len(lins) - 1)]
            _x = lins + start_region
            x_values = [float(_x[x] + _x[x + 1]) / 2 for x in range(len(lins) - 1)]
            self.ax.fill_between(x_values, scores_per_bin, linewidth=0.1,
                                 color=self.properties['color'],
                                 facecolor=self.properties['color'])

        else:
            # on rare occasions pyBigWig may throw an error, apparently caused by a corruption
            # of the memory. This only occurs when calling trackPlot from different
            # processors. Reloading the file solves the problem.
            num_tries = 0
            while num_tries < 5:
                num_tries += 1
                try:
                    scores = np.array(self.bw.stats(chrom_region, start_region, end_region, nBins=num_bins)).astype(float)
                except Exception as e:
                    import pyBigWig
                    self.bw = pyBigWig.open(self.properties['file'])

                    print "error found while reading bigwig scores ({}).\nTrying again. Iter num: {}".format(e, num_tries)
                    pass
                else:
                    if num_tries > 1:
                        print "After {} the scores could be computed".format(num_tries)
                    break

            x_values = np.linspace(start_region, end_region, num_bins)
            self.ax.fill_between(x_values, scores, linewidth=0.1,
                                 color=self.properties['color'],
                                 facecolor=self.properties['color'], zorder=1)
        self.ax.set_xlim(start_region, end_region)
        ymin, ymax = self.ax.get_ylim()
        if 'max_value' in self.properties and self.properties['max_value'] != 'auto':
            ymax = self.properties['max_value']
        if 'min_value' in self.properties and self.properties['min_value'] != 'auto':
            ymin = self.properties['min_value']

        if 'orientation' in self.properties and self.properties['orientation'] == 'inverted':
            self.ax.set_ylim(ymax, ymin)
        else:
            self.ax.set_ylim(ymin, ymax)

    #    self.ax.set_yticks([ymax])
        ydelta = ymax - ymin

    #    self.ax.set_yticklabels(["{}-{}".format(int(ymin), int(ymax))], size='large')
        # set min max
        if float(ymax) % 1 == 0:
            ymax_print = int(ymax)
        else:
            ymax_print = "{:.1f}".format(ymax)
        small_x = 0.01 * (end_region - start_region)
        if 'show data range' in self.properties and self.properties['show data range'] == 'no':
            pass
        else:
            # by default show the data range
            self.ax.text(start_region - small_x, ymax - ydelta * 0.2,
                         "[{}-{}]".format(int(ymin), ymax_print),
                         horizontalalignment='left',
                         verticalalignment='top')

        """
        self.ax.text(region_end, ymax - ydelta * 0.2, self.properties['title'],
                horizontalalignment='right', size='large',
                verticalalignment='bottom')

        """
        self.label_ax.text(0.15, 0.5, self.properties['title'],
                           horizontalalignment='left', size='large',
                           verticalalignment='center')

        return self.ax


class PlotHiCMatrix(TrackPlot):

    def __init__(self, properties_dict):
        # to avoid the color bar to span all the
        # width of the axis I pass two axes
        # to plot_matrix
        self.properties = properties_dict

        self.hic_ma = HiCMatrix.hiCMatrix(self.properties['file'])
        if len(self.hic_ma.matrix.data) == 0:
            exit("Matrix {} is empty".format(self.properties['file']))
        if 'show_masked_bins' in self.properties and self.properties['show_masked_bins'] == 'yes':
            pass
        else:
            self.hic_ma.maskBins(self.hic_ma.nan_bins)

        # check that the matrix can be log transformed
        if 'transform' in self.properties:
            if self.properties['transform'] == 'log1p':
                if self.hic_ma.matrix.data.min() + 1 < 0:
                    exit("\n*ERROR*\nMatrix contains negative values.\n"
                         "log1p transformation can not be applied to \n"
                         "values in matrix: {}".format(self.properties['file']))

            elif self.properties['transform'] == '-log':
                if self.hic_ma.matrix.data.min() < 0:
                    exit("\n*ERROR*\nMatrix contains negative values.\n"
                         "log(-1 * <values>) transformation can not be applied to \n"
                         "values in matrix: {}".format(self.properties['file']))

            elif self.properties['transform'] == 'log':
                if self.hic_ma.matrix.data.min() < 0:
                    exit("\n*ERROR*\nMatrix contains negative values.\n"
                         "log transformation can not be applied to \n"
                         "values in matrix: {}".format(self.properties['file']))

        new_intervals = hicexplorer.utilities.enlarge_bins(self.hic_ma.cut_intervals)
        self.hic_ma.interval_trees, self.hic_ma.chrBinBoundaries = \
            self.hic_ma.intervalListToIntervalTree(new_intervals)

        self.hic_ma.cut_intervals = new_intervals
        binsize = self.hic_ma.getBinSize()
        max_depth_in_bins = int(self.properties['depth'] / binsize)

        # work only with the lower matrix
        # and remove all pixels that are beyond
        # 2 * max_depth_in_bis which are not required
        # (this is done by subtracting a second sparse matrix
        # that contains only the lower matrix that wants to be removed.
        limit = 2 * max_depth_in_bins
        self.hic_ma.matrix = scipy.sparse.triu(self.hic_ma.matrix, k=0, format='csr') - \
            scipy.sparse.triu(self.hic_ma.matrix, k=limit, format='csr')
        self.hic_ma.matrix.eliminate_zeros()

        # fill the main diagonal, otherwise it looks
        # not so good. The main diagonal is filled
        # with an array containing the max value found
        # in the matrix
        if sum(self.hic_ma.matrix.diagonal()) == 0:
            sys.stderr.write("Filling main diagonal because is empty and "
                             "otherwise it looks bad...\n")
            max_value = self.hic_ma.matrix.data.max()
            main_diagonal = scipy.sparse.dia_matrix(([max_value] * self.hic_ma.matrix.shape[0], [0]),
                                                    shape=self.hic_ma.matrix.shape)
            self.hic_ma.matrix = self.hic_ma.matrix + main_diagonal

        self.plot_inverted = False
        if 'orientation' in self.properties and self.properties['orientation'] == 'inverted':
            self.plot_inverted = True

        self.norm = None

        if 'colormap' not in self.properties:
            self.properties['colormap'] = DEFAULT_MATRIX_COLORMAP

        self.cmap = matplotlib.cm.get_cmap(self.properties['colormap'])
        self.cmap.set_bad('white')

        self.cmap.set_bad('black')

        if 'boundaries_file' in self.properties:
            self.boundaries_obj = PlotBoundaries({'file': self.properties['boundaries_file']})

    def plot(self, ax, label_ax, chrom, region_start, region_end):
        import copy
        self.cbar_ax = copy.copy(label_ax)
        self.label_ax = label_ax
        self.label_ax.set_axis_off()
        self.ax = ax

        chrom_sizes = self.hic_ma.get_chromosome_sizes()
        if chrom not in chrom_sizes.keys():
            chrom = change_chrom_names(chrom)
        if region_end > chrom_sizes[chrom]:
            sys.stderr.write("*Error*\nThe region to plot extends beyond the chromosome size. Please check.\n")
            sys.stderr.write("{} size: {}. Region to plot {}-{}\n".format(chrom, chrom_sizes[chrom],
                                                                          region_start, region_end))

        # expand region to plus depth on both sides
        # to avoid a 45 degree 'cut' on the edges

        # get bin id of start and end of region in given chromosome
        chr_start_id, chr_end_id = self.hic_ma.getChrBinRange(chrom)
        chr_start = self.hic_ma.cut_intervals[chr_start_id][1]
        chr_end = self.hic_ma.cut_intervals[chr_end_id - 1][1]
        start_bp = max(chr_start, region_start - self.properties['depth'])
        end_bp = min(chr_end, region_end + self.properties['depth'])

        idx, start_pos = zip(*[(idx, x[1]) for idx, x in
                               enumerate(self.hic_ma.cut_intervals)
                               if x[0] == chrom and x[1] >= start_bp and x[2] <= end_bp])

        idx = idx[0:-1]
        # select only relevant matrix part
        matrix = self.hic_ma.matrix[idx, :][:, idx]
        # limit the 'depth' based on the length of the region being viewed

        region_len = region_end - region_start
        depth = min(self.properties['depth'], int(region_len * 1.25))
        depth_in_bins = int(1.5 * region_len / self.hic_ma.getBinSize())

        if depth < self.properties['depth']:
            # remove from matrix all data points that are not visible.
            matrix = matrix - scipy.sparse.triu(matrix, k=depth_in_bins, format='csr')
        matrix = np.asarray(matrix.todense().astype(float))

        if 'transform' in self.properties:
            if self.properties['transform'] == 'log1p':
                matrix += 1
                self.norm = matplotlib.colors.LogNorm()

            elif self.properties['transform'] == '-log':
                mask = matrix == 0
                matrix[mask] = matrix[mask is False].min()
                matrix = -1 * np.log(matrix)

            elif self.properties['transform'] == 'log':
                mask = matrix == 0
                matrix[mask] = matrix[mask is False].min()
                matrix = np.log(matrix)

        if 'max_value' in self.properties and self.properties['max_value'] != 'auto':
            vmax = self.properties['max_value']

        else:
            # try to use a 'aesthetically pleasant' max value
            vmax = np.percentile(matrix.diagonal(1), 80)

        if 'min_value' in self.properties and self.properties['min_value'] != 'auto':
            vmin = self.properties['min_value']
        else:
            if depth_in_bins > matrix.shape[0]:
                depth_in_bins = matrix.shape[0] - 5
            vmin = np.median(matrix.diagonal(int(region_len / self.hic_ma.getBinSize())))

        sys.stderr.write("setting min, max values for track {} to: {}, {}\n".format(self.properties['section_name'],
                                                                                    vmin, vmax))
        img = self.pcolormesh_45deg(matrix, start_pos, vmax=vmax, vmin=vmin)
        img.set_rasterized(True)
        if self.plot_inverted:
            self.ax.set_ylim(depth, 0)
        else:
            self.ax.set_ylim(0, depth)

        # ##plot boundaries
        # if a boundaries file is given, plot the
        # tad boundaries as line delineating the TAD triangles
        if 'boundaries_file' in self.properties:
            self.boundaries_obj.plot(ax, label_ax, chrom, region_start, region_end)

        self.ax.set_xlim(region_start, region_end)
        if 'x labels' in self.properties and self.properties['x labels'] != 'no':
            ticks = self.ax.get_xticks()
            labels = ["{:.2f}".format((x / 1e6))
                      for x in ticks]
            labels[-1] += "Mbp"
            self.ax.get_xaxis().set_tick_params(
                which='both',
                bottom='on',
                top='off',
                direction='out')

            self.ax.set_xticklabels(labels)
        else:
            self.ax.get_xaxis().set_tick_params(
                which='both',
                bottom='off',
                top='off',
                direction='out')
            self.ax.axes.get_xaxis().set_visible(False)

        self.ax.set_frame_on(False)
        self.ax.axes.get_yaxis().set_visible(False)
        self.cbar_ax.patch.set_alpha(0.0)
        try:
            if 'transform' in self.properties and \
                    self.properties['transform'] in ['log', 'log1p']:
                from matplotlib.ticker import LogFormatter
                formatter = LogFormatter(10, labelOnlyBase=False)
                # get a useful log scale
                # that looks like [1, 2, 5, 10, 20, 50, 100, ... etc]
                aa = np.array([1, 2, 5])
                tick_values = np.concatenate([aa * 10**x for x in range(10)])
                cobar = plt.colorbar(img, ticks=tick_values, format=formatter, ax=self.cbar_ax, fraction=0.95)
            else:
                cobar = plt.colorbar(img, ax=self.cbar_ax, fraction=0.95)
            cobar.solids.set_edgecolor("face")
            cobar.ax.set_ylabel(self.properties['title'])

            # adjust the labels of the colorbar
            labels = cobar.ax.get_yticklabels()
            ticks = cobar.ax.get_yticks()
            if ticks[0] == 0:
                # if the label is at the start of the colobar
                # move it above avoid being cut or overlapping with other track
                labels[0].set_verticalalignment('bottom')
            if ticks[-1] == 1:
                # if the label is at the end of the colobar
                # move it a bit inside to avoid overlapping
                # with other labels
                labels[-1].set_verticalalignment('top')
            cobar.ax.set_yticklabels(labels)

        except ValueError:
            pass
        # self.label_ax.text(0.3, 0.0, self.properties['title'], rotation=90,
        #                   horizontalalignment='left', size='large',
        #                   verticalalignment='bottom', transform=self.label_ax.transAxes)

    def pcolormesh_45deg(self, matrix_c, start_pos_vector, vmin=None,
                         vmax=None):
        """
        Turns the matrix 45 degrees and adjusts the
        bins to match the actual start end positions.
        """
        import itertools
        # code for rotating the image 45 degrees
        n = matrix_c.shape[0]
        # create rotation/scaling matrix
        t = np.array([[1, 0.5], [-1, 0.5]])
        # create coordinate matrix and transform it
        matrix_a = np.dot(np.array([(i[1], i[0])
                                    for i in itertools.product(start_pos_vector[::-1],
                                                               start_pos_vector)]), t)
        # this is to convert the indices into bp ranges
        x = matrix_a[:, 1].reshape(n + 1, n + 1)
        y = matrix_a[:, 0].reshape(n + 1, n + 1)
        # plot
        im = self.ax.pcolormesh(x, y, np.flipud(matrix_c),
                                vmin=vmin, vmax=vmax, cmap=self.cmap, norm=self.norm)
        return im


class PlotXAxis(TrackPlot):

    def __init__(self, *args, **kwargs):
        super(PlotXAxis, self).__init__(*args, **kwargs)
        if 'fontsize' not in self.properties:
            self.properties['fontsize'] = 15

    def plot(self, ax, label_axis, chrom_region, region_start, region_end):
        ax.set_xlim(region_start, region_end)
        ticks = ax.get_xticks()
        if ticks[-1] - ticks[1] <= 1e5:
            labels = ["{:.0f} kb".format((x / 1e3))
                      for x in ticks]

        elif 1e5 < ticks[-1] - ticks[1] < 4e6:
            labels = ["{:,.0f} kb".format((x / 1e3))
                      for x in ticks]
        else:
            labels = ["{:,.1f} Mbp".format((x / 1e6))
                      for x in ticks]
            # labels[-1] += "Mbp"

        ax.axis["x"] = ax.new_floating_axis(0, 0.5)

        ax.axis["x"].axis.set_ticklabels(labels)
        ax.axis['x'].axis.set_tick_params(which='minor', bottom='on')

        ax.axis["x"].major_ticklabels.set(size=int(self.properties['fontsize']))

        if 'where' in self.properties and self.properties['where'] == 'top':
            ax.axis["x"].set_axis_direction("top")


class PlotBoundaries(TrackPlot):

    def __init__(self, *args, **kwargs):
        super(PlotBoundaries, self).__init__(*args, **kwargs)

        line_number = 0
        interval_tree = {}
        intervals = []
        prev_chrom = None
        valid_intervals = 0

        with open(self.properties['file'], 'r') as file_h:
            for line in file_h.readlines():
                line_number += 1
                if line.startswith('browser') or line.startswith('track') or line.startswith('#'):
                    continue
                try:
                    chrom, start, end = line.strip().split('\t')[0:3]
                except Exception as detail:
                    msg = 'Could not read line\n{}\n. {}'.format(line, detail)
                    sys.exit(msg)

                try:
                    start = int(start)
                    end = int(end)
                except ValueError as detail:
                    msg = "Error reading line: {}. One of the fields is not " \
                          "an integer.\nError message: {}".format(line_number, detail)
                    sys.exit(msg)

                assert start <= end, "Error in line #{}, end1 larger than start1 in {}".format(line_number, line)

                if prev_chrom and chrom != prev_chrom:
                    start_array, end_array = zip(*intervals)
                    start_array = np.array(start_array)
                    end_array = np.array(end_array)
                    # check if intervals are consecutive or 1bp positions demarcating the boundaries
                    if np.any(end_array - start_array == 1):
                        # The file contains only boundaries at 1bp position.
                        end_array = start_array[1:]
                        start_array = start_array[:-1]
                    interval_tree[prev_chrom] = IntervalTree()
                    for idx in range(len(start_array)):
                        interval_tree[prev_chrom].add(Interval(start_array[idx], end_array[idx]))
                        valid_intervals += 1
                    intervals = []

                intervals.append((start, end))

                # each interval spans from the smallest start to the largest end
                prev_chrom = chrom

        start, end = zip(*intervals)
        start = np.array(start)
        end = np.array(end)
        # check if intervals are consecutive or 1bp positions demarcating the boundaries
        if np.any(end - start == 1):
            # The file contains only boundaries at 1bp position.
            end = start[1:]
            start = start[:-1]
        interval_tree[chrom] = IntervalTree()
        for idx in range(len(start)):
            interval_tree[chrom].add(Interval(start[idx], end[idx]))
            valid_intervals += 1

        if valid_intervals == 0:
            sys.stderr.write("No valid intervals were found in file {}".format(self.properties['file']))

        file_h.close()
        self.interval_tree = interval_tree

    def plot(self, ax, label_ax, chrom_region, start_region, end_region):
        """
        Plots the boundaries as triangles in the given ax.
        """
        x = []
        y = []
        if chrom_region not in self.interval_tree:
            orig = chrom_region
            chrom_region = change_chrom_names(chrom_region)
            print 'changing {} to {}'.format(orig, chrom_region)
        for region in sorted(self.interval_tree[chrom_region][start_region:end_region]):
            """
                  /\
                 /  \
                /    \
            _____________________
               x1 x2 x3
            """
            x1 = region.begin
            x2 = x1 + float(region.end - region.begin) / 2
            x3 = region.end
            y1 = 0
            y2 = (region.end - region.begin)
            x.extend([x1, x2, x3])
            y.extend([y1, y2, y1])

        ax.plot(x, y, color='black')
        ax.set_xlim(start_region, end_region)


class PlotBed(TrackPlot):

    def __init__(self, *args, **kwarg):
        super(PlotBed, self).__init__(*args, **kwarg)

        from matplotlib import font_manager
        if 'fontsize' not in self.properties:
            self.properties['fontsize'] = 12
        else:
            self.properties['fontsize'] = float(self.properties['fontsize'])

        self.fp = font_manager.FontProperties(size=self.properties['fontsize'])

        if 'color' not in self.properties:
            self.properties['color'] = DEFAULT_BED_COLOR
        if 'border_color' not in self.properties:
            self.properties['border_color'] = 'black'
        if 'labels' not in self.properties:
            self.properties['labels'] = 'on'
        if 'style' not in self.properties:
            self.properties['style'] = 'flybase'
        if 'display' not in self.properties:
            self.properties['display'] = 'stacked'
        if 'interval height' not in self.properties:
            self.properties['interval_height'] = 100

        # to set the distance between rows
        self.row_scale = self.properties['interval_height'] * 2.3

        self.colormap = None
        # check if the color given is a color map
        color_options = [m for m in matplotlib.cm.datad]
        if self.properties['color'] in color_options:
            norm = matplotlib.colors.Normalize(vmin=self.properties['min_value'],
                                               vmax=self.properties['max_value'])
            cmap = matplotlib.cm.get_cmap(self.properties['color'])
            self.colormap = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap)

        # self.process_bed()

    def process_bed(self, fig_width, region_start, region_end):

        # to improve the visualization of the genes
        # it is good to have an estimation of the label
        # length. In the following code I try to get
        # the length of a 'W'.
        if self.properties['labels'] == 'on':
            # from http://scipy-cookbook.readthedocs.org/items/Matplotlib_LaTeX_Examples.html
            inches_per_pt = 1.0 / 72.27
            font_in_inches = self.properties['fontsize'] * inches_per_pt
            region_len = region_end - region_start
            bp_per_inch = region_len / fig_width
            font_in_bp = font_in_inches * bp_per_inch
            self.len_w = font_in_bp
            print "len of w set to: {} bp".format(self.len_w)
        else:
            self.len_w = 1

        import readBed

        bed_file_h = readBed.ReadBed(opener(self.properties['file']))
        self.bed_type = bed_file_h.file_type
        valid_intervals = 0
        self.max_num_row = {}
        self.interval_tree = {}
        prev = None
        # check for the number of other intervals that overlap
        #    with the given interval
        #            1         2
        #  012345678901234567890123456
        #  1=========       4=========
        #       2=========
        #         3============
        #
        # for 1 row_last_position = [9]
        # for 2 row_last_position = [9, 14]
        # for 3 row_last_position = [9, 14, 19]
        # for 4 row_last_position = [26, 14, 19]

        row_last_position = []  # each entry in this list contains the end position
        # of genomic interval. The list index is the row
        # in which the genomic interval was plotted.
        # Any new genomic interval that wants to be plotted,
        # knows the row to use by finding the list index that
        # is larger than its start
        for bed in bed_file_h:
            if prev is not None:
                if prev.chromosome == bed.chromosome:
                    if bed.start < prev.start:
                        import ipdb
                        ipdb.set_trace()
                    assert bed.start >= prev.start, "*Error* Bed file not sorted. Please sort using sort -k1,1 -k2,2n"
                if prev.chromosome != bed.chromosome:
                    # init var
                    row_last_position = []
                    self.max_num_row[bed.chromosome] = 0
            else:
                self.max_num_row[bed.chromosome] = 0

            # check for overlapping genes including
            # label size (if plotted)

            if self.properties['labels'] == 'on':
                bed_extended_end = int(bed.end + (len(bed.name) * self.len_w))
            else:
                bed_extended_end = (bed.end + 2 * self.small_relative)

            # get smallest free row
            if len(row_last_position) == 0:
                free_row = 0
                row_last_position.append(bed_extended_end)
            else:
                # get list of rows that are less than bed.start, then take the min
                idx_list = [idx for idx, value in enumerate(row_last_position) if value < bed.start]
                if len(idx_list):
                    free_row = min(idx_list)
                    row_last_position[free_row] = bed_extended_end
                else:
                    free_row = len(row_last_position)
                    row_last_position.append(bed_extended_end)

            if bed.chromosome not in self.interval_tree:
                self.interval_tree[bed.chromosome] = IntervalTree()

            self.interval_tree[bed.chromosome].add(Interval(bed.start, bed.end, (bed, free_row)))
            valid_intervals += 1
            prev = bed
            if free_row > self.max_num_row[bed.chromosome]:
                self.max_num_row[bed.chromosome] = free_row

        print self.max_num_row

        if valid_intervals == 0:
            sys.stderr.write("No valid intervals were found in file {}".format(self.properties['file_name']))

    def get_y_pos(self, bed, free_row):
        """
        The y_pos is set such that regions to be plotted do not overlap (stacked). To override this
        the properties['collapsed'] needs to be set.

        The algorithm uses a interval tree (self.region_interval) to check the overlaps
        and a sort of coverage vector 'rows used' to identify the row in which to plot
        :return: int y position
        """

        # if the domain directive is given, ypos simply oscilates between 0 and 100
        if self.properties['display'] == 'interlaced':
            ypos = self.properties['interval_height'] if self.counter % 2 == 0 else 1

        elif self.properties['display'] == 'collapsed':
            ypos = 0

        else:
            ypos = free_row * self.row_scale
        return ypos

    def plot(self, ax, label_ax, chrom_region, start_region, end_region):
        self.counter = 0
        self.small_relative = 0.004 * (end_region - start_region)
        self.process_bed(ax.get_figure().get_figwidth(), start_region, end_region)

        if chrom_region not in self.interval_tree.keys():
            chrom_region = change_chrom_names(chrom_region)

        genes_overlap = sorted(self.interval_tree[chrom_region][start_region:end_region])

        # turn labels off when too many intervals are visible.
        if self.properties['labels'] != 'off' and len(genes_overlap) > 60:
            self.properties['labels'] = 'off'

        max_num_row_local = 1
        max_ypos = 0
        for region in genes_overlap:
            """
            BED12 gene format with exon locations at the end
            chrX    20850   23076   CG17636-RA      0       -       20850   23017   0       3       946,765,64,     0,1031,2162,

            BED9
            bed with rbg at end
            chr2L   0       70000   ID_5    0.26864549832   .       0       70000   51,160,44

            BED6
            bed with rbg at end
            chr2L   0       70000   ID_5    0.26864549832   .
            """
            self.counter += 1
            # an namedTuple object is stored in the region.value together with the free value for the gene
            bed, free_row = region.data
            if self.colormap:
                # translate value field (in the example above is 0 or 0.2686...) into a color
                rgb = self.colormap.to_rgba(bed.score)
                edgecolor = self.colormap.to_rgba(bed.score)
            else:
                rgb = self.properties['color']
                edgecolor = self.properties['border_color']

            # if rgb is set in the bed line, this overrides the previously
            # defined colormap
            if self.bed_type in ['bed9', 'bed12'] and len(bed.rgb) == 3:
                try:
                    rgb = [float(x) / 255 for x in bed.rgb]
                    if 'border_color' in self.properties:
                        edgecolor = self.properties['border_color']
                    else:
                        edgecolor = self.properties['color']
                except IndexError:
                    pass

            ypos = self.get_y_pos(bed, free_row)

            # do not plot if the maximum interval rows to plot is reached
            if 'gene rows' in self.properties and free_row >= int(self.properties['gene rows']):
                continue

            if free_row > max_num_row_local:
                max_num_row_local = free_row
            if ypos > max_ypos:
                max_ypos = ypos

            if self.bed_type == 'bed12':
                if self.properties['style'] == 'flybase':
                    self.draw_gene_with_introns_flybase_style(ax, bed, ypos, rgb, edgecolor)
                else:
                    self.draw_gene_with_introns(ax, bed, ypos, rgb, edgecolor)
            else:
                self.draw_gene_simple(ax, bed, ypos, rgb, edgecolor)

            if self.properties['labels'] == 'off':
                pass
            elif bed.start > start_region and bed.end < end_region:
                ax.text(bed.end + self.small_relative, ypos + (float(self.properties['interval_height']) / 2),
                        bed.name, horizontalalignment='left',
                        verticalalignment='center', fontproperties=self.fp)

        if self.counter == 0:
            sys.stderr.write("*Warning* No intervals were found for file {} \n"
                             "in section '{}' for the interval plotted ({}:{}-{}).\n"
                             "".format(self.properties['file'],
                                       self.properties['section_name'],
                                       chrom_region, start_region, end_region))

        ymax = 0

        if 'global max row' in self.properties and self.properties['global max row'] == 'yes':
            ymin = self.max_num_row[chrom_region] * self.row_scale

        elif 'gene rows' in self.properties:
            ymin = int(self.properties['gene rows']) * self.row_scale
        else:
            ymin = max_ypos + self.properties['interval_height']

        print("ylim {},{}".format(ymin, ymax))
        # the axis is inverted (thus, ymax < ymin)
        ax.set_ylim(ymin, ymax)

        """
        ax.text(start_region, (self.max_num_row[bed.chromosome] + 1) * 330 * 0.5,
                     "{}".format(self.properties['title']),
                     horizontalalignment='left', size='small',
                     verticalalignment='bottom')
        """

        if 'display' in self.properties:
            if self.properties['display'] == 'domain':
                ax.set_ylim(-5, 205)
            elif self.properties['display'] == 'collapsed':
                ax.set_ylim(-5, 105)

        ax.set_xlim(start_region, end_region)

        label_ax.text(0.15, 1, self.properties['title'],
                      horizontalalignment='left', size='large',
                      verticalalignment='top', transform=label_ax.transAxes)

    def draw_gene_simple(self, ax, bed, ypos, rgb, edgecolor):
        """
        draws an interval with direction (if given)
        """
        from matplotlib.patches import Polygon

        if bed.strand not in ['+', '-']:
            ax.add_patch(Rectangle((bed.start, ypos), bed.end - bed.start, self.properties['interval_height'],
                                   edgecolor=edgecolor, facecolor=rgb, linewidth=0.5))
        else:
            vertices = self._draw_arrow(ax, bed.start, bed.end, bed.strand, ypos)
            ax.add_patch(Polygon(vertices, closed=True, fill=True,
                                 edgecolor=edgecolor,
                                 facecolor=rgb,
                                 linewidth=0.5))

    def draw_gene_with_introns_flybase_style(self, ax, bed, ypos, rgb, edgecolor):
        """
        draws a gene using different styles
        """
        from matplotlib.patches import Polygon
        if bed.block_count == 0 and bed.thick_start == bed.start and bed.thick_end == bed.end:
            self.draw_gene_simple(ax, bed, ypos, rgb, edgecolor)
            return
        half_height = float(self.properties['interval_height']) / 2
        # draw 'backbone', a line from the start until the end of the gene
        ax.plot([bed.start, bed.end], [ypos + half_height, ypos + half_height], 'black', linewidth=0.5, zorder=-1)

        # get start, end of all the blocks
        positions = []
        for idx in range(0, bed.block_count):
            x0 = bed.start + bed.block_starts[idx]
            x1 = x0 + bed.block_sizes[idx]
            if x0 < bed.thick_start < x1:
                positions.append((x0, bed.thick_start, 'UTR'))
                positions.append((bed.thick_start, x1, 'coding'))

            elif x0 < bed.thick_end < x1:
                positions.append((x0, bed.thick_end, 'coding'))
                positions.append((bed.thick_end, x1, 'UTR'))

            else:
                if x1 < bed.thick_start or x0 > bed.thick_end:
                    type = 'UTR'
                else:
                    type = 'coding'

                positions.append((x0, x1, type))

        # plot all blocks as rectangles except the last if the strand is + or
        # the first is the strand is -, which are drawn as arrows.
        if bed.strand == '-':
            positions = positions[::-1]

        first_pos = positions.pop()
        if first_pos[2] == 'UTR':
            _rgb = 'grey'
        else:
            _rgb = rgb

        vertices = self._draw_arrow(ax, first_pos[0], first_pos[1], bed.strand, ypos)

        ax.add_patch(Polygon(vertices, closed=True, fill=True,
                             edgecolor=edgecolor,
                             facecolor=_rgb,
                             linewidth=0.5))

        for start_pos, end_pos, _type in positions:
            if _type == 'UTR':
                _rgb = 'grey'
            else:
                _rgb = rgb
            vertices = [(start_pos, ypos), (start_pos, ypos + self.properties['interval_height']),
                        (end_pos, ypos + self.properties['interval_height']), (end_pos, ypos)]

            ax.add_patch(Polygon(vertices, closed=True, fill=True,
                                 edgecolor=edgecolor,
                                 facecolor=_rgb,
                                 linewidth=0.5))

    def _draw_arrow(self, ax, start, end, strand, ypos):
        """
        Draws a filled arrow
        :param ax:
        :param start:
        :param end:
        :param strand:
        :param ypos:
        :param rgb:
        :return: None
        """
        half_height = float(self.properties['interval_height']) / 2
        if strand == '+':
            x0 = start
            x1 = end  # - self.small_relative
            y0 = ypos
            y1 = ypos + self.properties['interval_height']
            """
            The vertices correspond to 5 points along the path of a form like the following,
            starting in the lower left corner and progressing in a clock wise manner.

            -----------------\
            ---------------- /

            """

            vertices = [(x0, y0), (x0, y1), (x1, y1), (x1 + self.small_relative, y0 + half_height), (x1, y0)]

        else:
            x0 = start  # + self.small_relative
            x1 = end
            y0 = ypos
            y1 = ypos + self.properties['interval_height']
            """
            The vertices correspond to 5 points along the path of a form like the following,
            starting in the lower left corner and progressing in a clock wise manner.

            /-----------------
            \-----------------
            """
            vertices = [(x0, y0), (x0 - self.small_relative, y0 + half_height), (x0, y1), (x1, y1), (x1, y0)]

        return vertices

    def draw_gene_with_introns(self, ax, bed, ypos, rgb, edgecolor):
        """
        draws a gene like in flybase gbrowse.
        """
        from matplotlib.patches import Polygon

        if bed.block_count == 0 and bed.thick_start == bed.start and bed.thick_end == bed.end:
            self.draw_gene_simple(ax, bed, ypos, rgb, edgecolor)
            return
        half_height = float(self.properties['interval_height']) / 2
        quarter_height = float(self.properties['interval_height']) / 4
        three_quarter_height = quarter_height * 3

        # draw 'backbone', a line from the start until the end of the gene
        ax.plot([bed.start, bed.end], [ypos + half_height, ypos + half_height], 'black', linewidth=0.5, zorder=-1)

        for idx in range(0, bed.block_count):
            x0 = bed.start + bed.block_starts[idx]
            x1 = x0 + bed.block_sizes[idx]
            if x1 < bed.thick_start or x0 > bed.thick_end:
                y0 = ypos + quarter_height
                y1 = ypos + three_quarter_height
            else:
                y0 = ypos
                y1 = ypos + self.properties['interval_height']

            if x0 < bed.thick_start < x1:
                vertices = ([(x0, ypos + quarter_height), (x0, ypos + three_quarter_height),
                             (bed.thick_start, ypos + three_quarter_height),
                             (bed.thick_start, ypos + self.properties['interval_height']),
                             (bed.thick_start, ypos + self.properties['interval_height']),
                             (x1, ypos + self.properties['interval_height']), (x1, ypos),
                             (bed.thick_start, ypos), (bed.thick_start, ypos + quarter_height)])

            elif x0 < bed.thick_end < x1:
                vertices = ([(x0, ypos),
                             (x0, ypos + self.properties['interval_height']),
                             (bed.thick_end, ypos + self.properties['interval_height']),
                             (bed.thick_end, ypos + three_quarter_height),
                             (x1, ypos + three_quarter_height),
                             (x1, ypos + quarter_height),
                             (bed.thick_end, ypos + quarter_height),
                             (bed.thick_end, ypos)])
            else:
                vertices = ([(x0, y0), (x0, y1), (x1, y1), (x1, y0)])

            ax.add_patch(Polygon(vertices, closed=True, fill=True,
                                 linewidth=0.1,
                                 edgecolor='none',
                                 facecolor=rgb))

            if idx < bed.block_count - 1:
                # plot small arrows using the character '<' or '>' over the back bone
                intron_length = bed.block_starts[idx + 1] - (bed.block_starts[idx] + bed.block_sizes[idx])
                marker = 5 if bed.strand == '+' else 4
                if intron_length > 3 * self.small_relative:
                    pos = np.arange(x1 + 1 * self.small_relative,
                                    x1 + intron_length + self.small_relative, int(2 * self.small_relative))
                    ax.plot(pos, np.zeros(len(pos)) + ypos + half_height, '.', marker=marker,
                            fillstyle='none', color='blue', markersize=3)

                elif intron_length > self.small_relative:
                    intron_center = x1 + int(intron_length) / 2
                    ax.plot([intron_center], [ypos + half_height], '.', marker=5,
                            fillstyle='none', color='blue', markersize=3)


class PlotArcs(TrackPlot):

    def __init__(self, *args, **kwarg):
        import collections
        super(PlotArcs, self).__init__(*args, **kwarg)
        # the file format expected is similar to file format of links in
        # circos:
        # chr1 100 200 chr1 250 300 0.5
        # where the last value is an score.

        valid_intervals = 0
        interval_tree = {}
        line_number = 0
        Arc = collections.namedtuple("Arc", ['chrom1', 'start1', 'end1', 'start2', 'end2', 'score'])
        with open(self.properties['file'], 'r') as file_h:
            for line in file_h.readlines():
                line_number += 1
                if line.startswith('browser') or line.startswith('track') or line.startswith('#'):
                    continue
                try:
                    chrom1, start1, end1, chrom2, start2, end2, score = line.strip().split('\t')
                except Exception as detail:
                    msg = 'File not valid. The format is chrom1 start1, end1, ' \
                          'chrom2, start2, end2, score\nError: {}\n in line\n {}'.format(detail, line)
                    sys.exit(msg)

                try:
                    start1 = int(start1)
                    end1 = int(end1)
                    start2 = int(start2)
                    end2 = int(end2)
                except ValueError as detail:
                    msg = "Error reading line: {}. One of the fields is not " \
                          "an integer.\nError message: {}".format(line_number, detail)
                    sys.exit(msg)

                assert start1 <= end1, "Error in line #{}, end1 larger than start1 in {}".format(line_number, line)
                assert start2 <= end2, "Error in line #{}, end2 larger than start2 in {}".format(line_number, line)
                try:
                    score = float(score)
                except ValueError as detail:
                    msg = "Error reading line: {}. The score is not valid {}. " \
                          "\nError message: {}".format(line_number, detail)
                    sys.exit(msg)

                if chrom1 != chrom2:
                    sys.stderr.write("Only links in same chromosome are considere. Skipping line\n{}\n".format(line))
                    continue

                if chrom1 not in interval_tree:
                    interval_tree[chrom1] = IntervalTree()

                if start2 < start1:
                    start1, start2 = start2, start1
                    end1, end2 = end2, end1

                value = Arc._make([chrom1, start1, end1, start2, end2, score])

                # each interval spans from the smallest start to the largest end
                interval_tree[chrom1].add(Interval(start1, end2, value))
                valid_intervals += 1

        if valid_intervals == 0:
            sys.stderr.write("No valid intervals were found in file {}".format(self.properties['file']))

        file_h.close()
        self.interval_tree = interval_tree

        if 'color' not in self.properties:
            self.properties['color'] = 'blue'

        if 'alpha' not in self.properties:
            self.properties['alpha'] = 0.8

    def plot(self, ax, label_ax, chrom_region, region_start, region_end):
        """
        Makes and arc connecting two points on a linear scale representing
        interactions between Hi-C bins.
        :param ax: matplotlib axis
        :param label_ax: matplotlib axis for labels
        """
        from matplotlib.patches import Arc
        max_diameter = 0
        count = 0

        if chrom_region not in self.interval_tree.keys():
            chrom_region = change_chrom_names(chrom_region)
        arcs_in_region = sorted(self.interval_tree[chrom_region][region_start:region_end])

        for idx, interval in enumerate(arcs_in_region):
            # skip arcs whose start and end are outside the ploted region
            if interval.start < region_start and interval.end > region_end:
                continue

            diameter = interval.end - interval.start
            center = interval.start + float(diameter) / 2
            if diameter > max_diameter:
                max_diameter = diameter
            count += 1
            ax.plot([center], [diameter])
            if 'line width' in self.properties:
                line_width = float(self.properties['line width'])
            else:
                line_width = 0.5 * np.sqrt(interval.value.score)
            ax.add_patch(Arc((center, 0), diameter,
                             diameter * 2, 0, 0, 180, color=self.properties['color'], lw=line_width))

        print "{} arcs plotted".format(count)
        if 'orientation' in self.properties and self.properties['orientation'] == 'inverted':
            ax.set_ylim(max_diameter, -1)
        else:
            ax.set_ylim(-1, max_diameter)

        ax.set_xlim(region_start, region_end)
        label_ax.text(0.3, 0.0, self.properties['title'],
                      horizontalalignment='left', size='large',
                      verticalalignment='bottom', transform=label_ax.transAxes)


def change_chrom_names(chrom):
    """
    Changes UCSC chromosome names to ensembl chromosome names
    and vice versa.
    """
    # TODO: mapping from chromosome names like mithocondria is missing
    if chrom.startswith('chr'):
        # remove the chr part from chromosome name
        chrom = chrom[3:]
    else:
        # prefix with 'chr' the chromosome name
        chrom = 'chr' + chrom

    return chrom



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division
import sys
import logging
import argparse
import json
from hicexplorer import HiCMatrix as hm
from hicexplorer.utilities import enlarge_bins
from scipy import sparse
import numpy as np
import multiprocessing
from hicexplorer._version import __version__

logging.basicConfig()
log = logging.getLogger("hicFindTADs")
log.setLevel(logging.INFO)


def parse_arguments(args=None):
    """
    get command line arguments
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        conflict_handler='resolve',
        description="""
Uses a measure called TAD-separation score to identify the degree of separation between
the left and right regions at each Hi-C matrix bin. This is done for a
running window of different sizes. Then, TADs are called as those
positions having a local TAD-separation score minimum. The TAD-separation score is
measured using the z-score of the Hi-C matrix and is defined as the mean zscore of all
the matrix contacts between the left and right regions (diamond).

To find the TADs, the program  needs to compute first the
TAD scores at different window sizes. Then, the results of that computation
are used to call the TADs. This is convenient to test different filtering criteria quickly
as the demanding step is the computation of TAD-separation scores.

 An simple example usage is:

$ hicFindTads TAD_score -m hic_matrix.h5 -o TAD_score.txt

$ hicFindTads find_TADs -f TAD_score.txt --outPrefix TADs

For detailed help:

 hicFindTADs TAD_score -h

 or

 hicFindTADs find_TADs -h

""")

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    subparsers = parser.add_subparsers(
        title="commands",
        dest='command',
        metavar='')

    tad_score_subparser = subparsers.add_parser('TAD_score',
                                                formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # define the arguments
    tad_score_subparser.add_argument('--matrix', '-m',
                                     help='Corrected Hi-C matrix to use for the computations',
                                     required=True)

    tad_score_subparser.add_argument('--outFileName', '-o',
                                     help='File name to store the computation of the TAD-separation score. The format'
                                          'of the output file is chrom start end TAD-sep1 TAD-sep2 TAD-sep3 .. etc. '
                                          'We call this format a bedgraph matrix and can be plotted using '
                                          '`hicPlotTADs`. Each of the TAD-separation scores in the file corresponds to '
                                          'a different window length starting from --minDepth to --maxDepth.',
                                     required=True)

    tad_score_subparser.add_argument('--minDepth',
                                     help='Minimum window length (in bp) to be considered to the left and to the right '
                                          'of each Hi-C bin. This number should be at least 3 times '
                                          'as large as the bin size of the Hi-C matrix.',
                                     metavar='INT bp',
                                     type=int)

    tad_score_subparser.add_argument('--maxDepth',
                                     help='Maximum window length to be considered to the left and to the right '
                                          'of the cut point in bp. This number should around 6-10 times '
                                          'as large as the bin size of the Hi-C matrix.',
                                     metavar='INT bp',
                                     type=int)

    tad_score_subparser.add_argument('--step',
                                     help='Step size when moving from --minDepth to --maxDepth. Note, the step size'
                                          'grows exponentially as '
                                          '`maxDeph + (step * int(x)**1.5) for x in [0, 1, ...]` until  it '
                                          'reaches `maxDepth`. For example, selecting  step=10,000, minDepth=20,000 '
                                          'and maxDepth=150,000 will compute TAD-scores for window sizes: '
                                          '20,000, 30,000, 40,000, 70,000 and 100,000',
                                     metavar='INT bp',
                                     type=int)

    tad_score_subparser.add_argument('--numberOfProcessors', '-p',
                                     help='Number of processors to use ',
                                     type=int,
                                     default=1)

    find_tads_subparser = subparsers.add_parser('find_TADs', formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    find_tads_subparser.add_argument('--tadScoreFile', '-f',
                                     help='File containing the TAD scores (generated by running hicFindTADs TAD_score)',
                                     required=True)

    find_tads_subparser.add_argument('--minBoundaryDistance',
                                     help='Minimum distance between boundaries (in bp). This parameter can be '
                                          'used to reduce spurious boundaries caused by noise. ',
                                     type=int)

    find_tads_subparser.add_argument('--pvalue',
                                     help='P-value threshold. The probability of a local minima to be a boundary '
                                          'is estimated by comparing the distribution (Wilcoxon ranksum) of '
                                          'the  zscores between the left and right '
                                          'regions (diamond) at the local minimum with the matrix zscores for a '
                                          'diamond at --minDepth to the left and a diamond --minDepth to the right. '
                                          'The reported pvalue is the Bonferroni correction all pvalues. Default is '
                                          '0.01',
                                     type=float,
                                     default=0.01
                                     )

    find_tads_subparser.add_argument('--delta',
                                     help='Minimum threshold of the difference between the TAD-separation score of a '
                                          'putative boundary and the mean of the TAD-sep. score of surrounding bins. '
                                          'The delta value reduces spurious boundaries that are shallow, which usually '
                                          'occur at the center of large TADs when the TAD-sep. score is flat. Higher '
                                          'delta threshold values produce more conservative boundary estimations. By '
                                          'default a value of 0.01 is used.',
                                     type=float,
                                     default=0.01
                                     )

    find_tads_subparser.add_argument('--outPrefix',
                                     help='File prefix to save the resulting files: 1. <prefix>_boundaries.bed, which '
                                          'contains the positions of boundaries. The genomic coordinates in this file '
                                          'correspond to the resolution used. Thus, for Hi-C bins of '
                                          '10.000bp the boundary position is 10.000bp long. For restriction fragment '
                                          'matrices the boundary position varies depending on the fragment length '
                                          'at the boundary. 2. <prefix>_domains.bed '
                                          'contains the TADs positions. This is a non-overlapping set of genomic '
                                          'positions. 3. <prefix>_boundaries.gff Similar to the boundaries bed file '
                                          'but with extra information (pvalue, delta). 4. <prefix>_score.bedgraph file '
                                          'contains the TAD-separation score '
                                          'measured at each Hi-C bin coordinate. Is useful to visualize in a genome '
                                          'browser. The delta and pvalue settings are saved as part of the name.',
                                     required=True)

    return parser


def get_cut_weight_by_bin_id(matrix, cut, depth, return_mean=False):
    """
    like get_cut_weight which is the 'diamond' representing the counts
    from a region -dept to +depth from the given bin position (cut):

    """
    if cut < 0 or cut > matrix.shape[0]:
        return None
    # the range [start:i] should have running window
    # length elements (i is excluded from the range)
    start = max(0, cut - depth)
    # same for range [i+1:end] (i is excluded from the range)
    end = min(matrix.shape[0], cut + depth)

    # the idea is to evaluate the interactions
    # between the upstream neighbors with the
    # down stream neighbors. In other words
    # the inter-domain interactions
    if return_mean is True:
        return matrix[start:cut, cut:end].mean()
    else:
        return matrix[start:cut, cut:end].todense().A1


def get_idx_of_bins_at_given_distance(hic_matrix, idx, window_len):
    """
    Returns the right and left indices that are at a given distance in bp
    from `idx`
    Parameters
    ----------
    idx reference index
    window_len distance in bp

    Returns
    -------
    tuple, with left and right bin indices
    """

    # find the bin that is closer to window_len
    # the range [start:i] should have running window
    # length elements (i is excluded from the range)
    chrom, cut_start, cut_end, _ = hic_matrix.getBinPos(idx)
    left_start = max(0, cut_start - window_len)
    left_idx = hic_matrix.getRegionBinRange(chrom, left_start, left_start + 1)[0]

    chr_end_pos = hic_matrix.get_chromosome_sizes()[chrom]
    right_end = min(chr_end_pos, cut_end + window_len) - 1
    right_idx = hic_matrix.getRegionBinRange(chrom, right_end, right_end)[0]

    return left_idx, right_idx


def get_cut_weight(hic_matrix, cut, window_len, return_mean=False):
    """
    like get_cut_weight which is the 'diamond' representing the counts
    from a region -window_len to +window_len from the given bin position (cut):

    """
    if cut < 0 or cut > hic_matrix.matrix.shape[0]:
        return None

    left_idx, right_idx = get_idx_of_bins_at_given_distance(hic_matrix, cut, window_len)

    if return_mean is True:
        return hic_matrix.matrix[left_idx:cut, cut:right_idx].mean()
    else:
        return hic_matrix.matrix[left_idx:cut, cut:right_idx].todense().A1


def get_triangle(hic_matrix, cut, window_len, return_mean=False):
    """
    like get_cut_weight which is the 'diamond' representing the counts
    from a region -window_len to +window_len from the given bin position (cut):

    """
    if cut < 0 or cut > hic_matrix.matrix.shape[0]:
        return None

    left_idx, right_idx = get_idx_of_bins_at_given_distance(hic_matrix, cut, window_len)

    def remove_lower_triangle(matrix):
        """
        remove all values in the lower triangle of a matrix
        """
        return matrix[np.triu_indices_from(matrix)].A1

    edges_left = remove_lower_triangle(hic_matrix.matrix[left_idx:cut, :][:, left_idx:cut].todense())
    edges_right = remove_lower_triangle(hic_matrix.matrix[cut:right_idx, :][:, cut:right_idx].todense())
#    if cut > 1000:
#        import ipdb;ipdb.set_trace()
    return np.concatenate([edges_left, edges_right])


def compute_matrix_wrapper(args):
    return compute_matrix(*args)


def get_incremental_step_size(min_win_size, max_win_size, start_step_len):
    """
    generates a list of incremental windows sizes (measured in bins)

    :param min_win_size: starting window size
    :param max_win_size: end window size
    :param start_step_len: start length
    :return: incremental_step list of bin lengths
    """
    incremental_step = []
    step = -1
    while 1:
        step += 1
        inc_step = min_win_size + int(start_step_len * (step**1.5))
        if step > 1 and inc_step == incremental_step[-1]:
            continue
        if inc_step > max_win_size:
            break
        incremental_step.append(inc_step)
    return incremental_step


def compute_matrix(bins_list, min_win_size=8, max_win_size=50, step_len=2):
    """
    Iterates over the Hi-C matrix computing at each bin
    interface the conductance at different window lengths
    :param hic_ma: Hi-C matrix object from HiCMatrix
    :param outfile: String, path of a file to save the conductance
                matrix in *bedgraph matrix* format
    :return: (chrom, start, end, matrix)
    """
    global hic_ma
    positions_array = []
    cond_matrix = []
    incremental_step = get_incremental_step_size(min_win_size, max_win_size, step_len)
    for cut in bins_list:

        chrom, chr_start, chr_end, _ = hic_ma.cut_intervals[cut]

        # get conductance
        # for multiple window lengths at a time
        mult_matrix = [get_cut_weight(hic_ma, cut, depth, return_mean=True) for depth in incremental_step]
        if np.any(np.isnan(mult_matrix)):
            # skip problematic cases
            continue

        cond_matrix.append(mult_matrix)

        positions_array.append((chrom, chr_start, chr_end))

    try:
        chrom, chr_start, chr_end = zip(*positions_array)
    except:
        import ipdb
        ipdb.set_trace()
    cond_matrix = np.vstack(cond_matrix)

    return chrom, chr_start, chr_end, cond_matrix


def peakdetect(y_axis, x_axis=None, lookahead=3, delta=0, chrom=None):
    """
    Based on the MATLAB script at:
    http://billauer.co.il/peakdet.html

    function for detecting local maximum and minimum in a signal.
    Discovers peaks by searching for values which are surrounded by lower
    or larger values for maximum and minimum respectively

    keyword arguments:
    :param: y_axis -- A list containig the signal over which to find peaks
    :param: x_axis -- (optional) A x-axis whose values correspond to the y_axis list
        and is used in the return to specify the position of the peaks. If
        omitted an index of the y_axis is used. (default: None)
    :param: lookahead -- (optional) distance to look ahead from a peak candidate to
        determine if it is the actual peak
    :param: delta -- (optional) this specifies a minimum difference between a peak and
        the following points, before a peak may be considered a peak. Useful
        to hinder the function from picking up false peaks towards to end of
        the signal. To work well delta should be set to delta >= RMSnoise * 5.


    :return: -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*tab)
    """
    max_peaks = []
    min_peaks = []
    dump = []   # Used to pop the first hit which almost always is false

    # check input data
    if x_axis is None:
        x_axis = np.arange(len(y_axis))

    if len(y_axis) != len(x_axis):
        raise (ValueError, 'Input vectors y_axis and x_axis must have same length')

    # store data length for later use

    if not (np.isscalar(delta) and delta >= 0):
        raise (ValueError, "delta must be a positive number")

    # maximum and minimum candidates are temporarily stored in
    # min_x and min_y respectively
    min_y, max_y = np.Inf, -np.Inf
    max_pos, min_pos = None, None
    search_for = None
    # Only detect peak if there is 'lookahead' amount of points after it
    prev_chrom = None
    for index, (x, y) in enumerate(zip(x_axis[:-lookahead], y_axis[:-lookahead])):
        assert -np.inf < y < np.inf, "Error, infinity value detected for value at position {}".format(index)
        if prev_chrom is None:
            prev_chrom = chrom[index]
        if prev_chrom != chrom[index]:
            # reset variables at the start of a chromosome
            min_y, max_y = np.Inf, -np.Inf
            max_pos, min_pos = None, None
            search_for = None

        prev_chrom = chrom[index]

        if y > max_y:
            max_y = y
            max_pos = x
        if y < min_y:
            min_y = y
            min_pos = x

        # look for max
        if y < max_y - delta and max_y != np.Inf and search_for != 'min':
            # Maximum peak candidate found
            # look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index + lookahead].max() < max_y:
                max_peaks.append([max_pos, max_y])
                dump.append(True)
                # set algorithm to only find minimum now
                max_y = y
                min_y = y
                min_pos = x
                search_for = 'min'
                continue

        # look for min
        if y > min_y + delta and min_y != -np.Inf and search_for != 'max':
            # Minimum peak candidate found
            # look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index + lookahead].min() > min_y:
                min_peaks.append([min_pos, min_y])
                dump.append(False)
                # set algorithm to only find maximum now
                min_y = y
                max_y = y
                max_pos = x
                search_for = 'max'

    # Remove the false hit on the first value of the y_axis
    try:
        if dump[0]:
            max_peaks.pop(0)
        else:
            min_peaks.pop(0)
        del dump
    except IndexError:
        # no peaks were found, should the function return empty lists?
        pass

    return [max_peaks, min_peaks]


def delta_wrt_window(min_idx_list, matrix_avg, chrom, window_len=10):
    """
    Computes the local 'delta' for each minima identified. The local
    delta is defined as the difference between the TAD-separation score at the minimum with respect
    to each of the TAD-separation scores of the `window_len` to the left, plus the `window_len` to
    the right of each minimum. The minimum itself is excluded.

    Parameters
    ----------
    matrix TAD score matrix for different window sizes
    chrom list of chrom names for each matrix row
    window_len the length of the window to look for the local TAD-score average

    Returns
    -------
    A dict of each minima delta to the mean of the TAD-separation score surrounding the minima. The keys
    are the min_idx

    """

    # compute the start and end points of the chromosomes
    # and store it as tuples in chrom_ranges list (e.g. chrom_ranges = [(0, 29254), (29254, 60006), ...]
    unique_chroms, chr_start_idx = np.unique(chrom, return_index=True)
    chr_start_idx = np.concatenate([chr_start_idx, [len(chrom) - 1]])
    chr_start_idx = np.sort(chr_start_idx)
    chrom_ranges = [(chr_start_idx[x], chr_start_idx[x + 1]) for x in range(len(chr_start_idx) - 1)]

    delta_to_mean = {}
    for min_idx in min_idx_list:
        # check that the min_idx is not to close to any of the chromosome boundaries
        close_to_chrom_border = True
        for start_range, end_range in chrom_ranges:
            if start_range < min_idx < end_range:
                if min_idx - window_len >= start_range and min_idx + window_len < end_range:
                    close_to_chrom_border = False
                    continue
        if close_to_chrom_border is True:
            delta_to_mean[min_idx] = np.nan

        else:
            local_tad_score = np.concatenate([matrix_avg[min_idx - window_len:min_idx + 3],
                                              matrix_avg[min_idx + 4: min_idx + window_len]])
            delta_to_mean[min_idx] = local_tad_score.mean() - matrix_avg[min_idx]

    return delta_to_mean


def find_consensus_minima(tad_score_matrix, lookahead=3, chrom=None):
    """
    Finds the minimum over the average values per column
    :param tad_score_matrix: TAD-separation score matrix
    :return:
    """

    tad_score_matrix_avg = tad_score_matrix.mean(axis=1)

    # compute local minima for the matrix average
    _max, _min = peakdetect(tad_score_matrix_avg, lookahead=lookahead, chrom=chrom)
    min_idx, value = zip(*_min)

    # get the delta for each boundary
    delta_to_mean = delta_wrt_window(min_idx, tad_score_matrix_avg, chrom)

    return min_idx, delta_to_mean


def hierarchical_clustering(boundary_list, clusters_cutoff=[]):
    """
    :param boundary_list: is a list of tuples each containing
    the location of a boundary. The order should be sorted
    and contain the following values:
        (chrom, start, value)
    :param clusters_cutoff: List of values to separate clusters. The
        clusters found at those value thresholds are returned.


    :return: z_value, clusters

    For z_value, the format used is similar as the scipy.cluster.hierarchy.linkage() function
    which is described as follows:

    A 4 by :math:`(n-1)` matrix ``z_value`` is returned. At the
    :math:`i`-th iteration, clusters with indices ``z_value[i, 0]`` and
    ``z_value[i, 1]`` are combined to form cluster :math:`n + i`. A
    cluster with an index less than :math:`n` corresponds to one of
    the :math:`n` original observations. The distance between
    clusters ``z_value[i, 0]`` and ``z_value[i, 1]`` is given by ``z_value[i, 2]``. The
    fourth value ``z_value[i, 3]`` represents the number of original
    observations in the newly formed cluster.

    The difference is that instead of a 4 times n-1 array, a
    6 times n-1 array is returned. Where positions 4, and 5
    correspond to the genomic coordinates of ``z_value[i, 0]`` and ``z_value[i, 1]``

    """
    # run the hierarchical clustering per chromosome
    if clusters_cutoff:
        # sort in reverse order
        clusters_cutoff = np.sort(np.unique(clusters_cutoff))[::-1]

    chrom, start, value = zip(*boundary_list)

    unique_chr, indices = np.unique(chrom, return_index=True)
    indices = indices[1:]  # the first element is not needed
    start_per_chr = np.split(start, indices)
    value_per_chr = np.split(value, indices)
    z_value = {}

    def get_domain_positions(boundary_position):
        """
        returns for each boundary a start,end position
        corresponding to each TAD
        :param boundary_position: list of boundary chromosomal positions
        :return: list of (start, end) tuples.
        """
        start_ = None
        domain_list = []
        for position in boundary_position:
            if start_ is None:
                start_ = position
                continue
            domain_list.append((start_, position))
            start_ = position

        return domain_list

    def find_in_clusters(clusters_, search_id):
        """
        Given a list of clusters (each cluster defined as as set,
        the function returns the position in which an id is found
        :param clusters_:
        :param search_id:
        :return:
        """
        for set_idx, set_of_ids in enumerate(clusters_):
            if search_id in set_of_ids:
                return set_idx

    def cluster_to_regions(clusters_, chrom_name):
        """
        Transforms a list of sets of ids from the hierarchical
        clustering to genomic positions
        :param clusters_: cluster ids
        :param chrom_name: chromosome name
        :return: list of tuples with (chrom_name, start, end)

        Example:

        clusters = [set(1,2,3), set(4,5,10)]

        """
        start_list = []
        end_list = []
        for set_ in clusters_:
            if len(set_) == 0:
                continue

            # the ids in the sets are created in such a
            # that the min id is the one with the smaller start position
            start_list.append(domains[min(set_)][0])
            end_list.append(domains[max(set_)][1])

        start_list = np.array(start_list)
        end_list = np.array(end_list)
        order = np.argsort(start_list)

        return zip([chrom_name] * len(order), start_list[order], end_list[order])

    return_clusters = {}  # collects the genomic positions of the clusters per chromosome
    # The values are a list, one for each cutoff.
    for chrom_idx, chrom_name in enumerate(unique_chr):
        z_value[chrom_name] = []
        return_clusters[chrom_name] = []
        clust_cutoff = clusters_cutoff[:]
        domains = get_domain_positions(start_per_chr[chrom_idx])
        clusters = [{x} for x in range(len(domains))]

        # initialize the cluster_x with the genomic position of domain centers
        cluster_x = [int(d_start + float(d_end - d_start) / 2) for d_start, d_end in domains]
        # number of domains should be equal to the number of values minus 1
        assert len(domains) == len(value_per_chr[chrom_idx]) - 1, "error"

        """
        domain:id
             0            1               2            3
         |---------|---------------|----------------|----|
        values:id
         0         1               3                3    4
        values id after removing flanks
                   0               1                2
         """
        values = value_per_chr[chrom_idx][1:-1]  # remove flanking values that do not join TADs

        # from highest to lowest merge neighboring domains
        order = np.argsort(values)[::-1]
        for idx, order_idx in enumerate(order):
            if len(clust_cutoff) and idx + 1 < len(order) and \
                    values[order_idx] >= clust_cutoff[0] > values[order[idx + 1]]:
                clust_cutoff = clust_cutoff[1:]  # remove first element
                return_clusters[chrom_name].append(cluster_to_regions(clusters, chrom_name))
            # merge domains order_idx - 1 and order_idx
            left = find_in_clusters(clusters, order_idx)
            right = find_in_clusters(clusters, order_idx + 1)
            z_value[chrom_name].append((left, right, values[order_idx],
                                        len(clusters[left]) + len(clusters[right]),
                                        cluster_x[left], cluster_x[right]))

            # set as new cluster position the center between the two merged
            # clusters
            gen_dist = int(float(abs(cluster_x[left] - cluster_x[right])) / 2)
            cluster_x.append(min(cluster_x[left], cluster_x[right]) + gen_dist)

            clusters.append(clusters[left].union(clusters[right]))
            clusters[left] = set()
            clusters[right] = set()

    # convert return_clusters from a per chromosome dictionary to
    # a per cut_off dictionary merging all chromosomes in to one list.
    ret_ = {}  # dictionary to hold the clusters per cutoff. The key of
    # each item is the str(cutoff)

    for idx, cutoff in enumerate(clusters_cutoff):
        cutoff = str(cutoff)
        ret_[cutoff] = []
        for chr_name in return_clusters:
            try:
                ret_[cutoff].extend(return_clusters[chr_name][idx])
            except IndexError:
                pass

    return z_value, ret_


def save_linkage(Z, file_name):
    """

    :param Z: Z has a format similar to the scipy.cluster.linkage matrix (see function
                hierarchical_clustering).
    :param file_name: File name to save the results
    :return: None
    """

    try:
        file_h = open(file_name, 'w')
    except IOError:
        log.error("Can't save linkage file:\n{}".format(file_name))
        return

    count = 0
    for chrom, values in Z.iteritems():
        for id_a, id_b, distance, num_clusters, pos_a, pos_b in values:
            count += 1
            file_h.write('{}\t{}\t{}\tclust_{}'
                         '\t{}\t.\t{}\t{}\t{}\n'.format(chrom,
                                                        int(pos_a),
                                                        int(pos_b),
                                                        count,
                                                        distance,
                                                        id_a, id_b,
                                                        num_clusters))


def get_domains(boundary_list):
    """
    returns for each boundary a chrom, start,end position
    corresponding to each TAD
    :param boundary_position: list of boundary chromosomal positions
    :return: list of (chrom, start, end, value) tuples.
    """
    prev_start = None
    prev_chrom = boundary_list[0][0]
    domain_list = []
    for chrom, start, value in boundary_list:
        if start is None:
            prev_start = start
            prev_chrom = chrom
            continue
        if prev_chrom != chrom:
            prev_chrom = chrom
            prev_start = None
            continue
        domain_list.append((chrom, prev_start, start, value))
        prev_start = start
        prev_chrom = chrom

    return domain_list


def toBytes(s):
    """
    Like toString, but for functions requiring bytes in python3
    """
    if sys.version_info[0] == 2:
        return s
    if isinstance(s, bytes):
        return s
    if isinstance(s, str):
        return bytes(s, 'ascii')
    if isinstance(s, list):
        return [toBytes(x) for x in s]
    return s


def save_bedgraph_matrix(outfile, chrom, chr_start, chr_end, score_matrix, args):
    """
    Save matrix as chrom, start, end ,row, values separated by tab
    I call this a bedgraph matrix (bm)

    :param outfile: string file name
    :param chrom: list of chrom names
    :param chr_start: list of start positions
    :param chr_end: list of end positions
    :param score_matrix: list of lists
    :return: None
    """
    params = dict(args._get_kwargs())
    params['zscore_matrix'] = args.outFileName + "_zscore_matrix.h5"
    params_str = json.dumps(params, separators=(',', ':'))

    with open(outfile, 'w') as f:
        f.write(toBytes("#" + params_str + "\n"))
        for idx in range(len(chrom)):
            matrix_values = "\t".join(
                np.char.mod('%f', score_matrix[idx, :]))
            f.write("{}\t{}\t{}\t{}\n".format(chrom[idx], chr_start[idx],
                                              chr_end[idx], matrix_values))


def save_clusters(clusters, file_prefix):
    """

    :param clusters: is a dictionary whose key is the cut of used to create it.
                     the value is a list of tuples, each representing
                      a genomec interval as ('chr', start, end).
    :param file_prefix: file prefix to save the resulting bed files
    :return: list of file names created
    """
    for cutoff, intervals in clusters.iteritems():
        fileh = open("{}_{}.bed".format(file_prefix, cutoff), 'w')
        for chrom, start, end in intervals:
            fileh.write("{}\t{}\t{}\t.\t0\t.\n".format(chrom, start, end))


def save_domains_and_boundaries(chrom, chr_start, chr_end, matrix, min_idx,
                                delta_of_min, pvalue_of_min, args):

    # a boundary is added to the start and end of each chromosome
    # np.unique return index is used to quickly get
    # the indices at which the name of the chromosome changes (chrom, start, end should be  sorted)
    unique_chroms, chr_start_idx = np.unique(chrom, return_index=True)

    # the trick to get the start positions using np.unique only works when
    # more than one chromosome is present
    if len(unique_chroms) == 1:
        chr_start_idx = [0]
        chr_end_idx = [len(chrom) - 1]
    else:
        chr_end_idx = chr_start_idx
        # the indices for the end of the chromosomes
        # are the the start indices - 1, with the exception
        # of the idx == 0 that is transformed into the length
        # of the chromosome to get the idx for the end of the
        # last chromosome
        chr_end_idx[chr_end_idx == 0] = len(chrom)
        chr_end_idx -= 1

    # put all indices together and sort
    min_idx = np.sort(np.concatenate([chr_start_idx, chr_end_idx, min_idx]))

    mean_mat_all = matrix.mean(axis=1)

    prefix = args.outPrefix
    filtered_min_idx = []
    for idx in min_idx:
        # filter by delta and pvalue_thresholds
        if idx not in delta_of_min:
            delta_of_min[idx] = np.nan
        if delta_of_min[idx] >= args.delta and pvalue_of_min[idx] <= args.pvalue:
            filtered_min_idx += [idx]
    log.info("Number of boundaries for delta {} and pval {}: {}".format(args.delta, args.pvalue,
                                                                        len(filtered_min_idx)))
    count = 1
    with open(prefix + '_boundaries.bed', 'w') as file_boundary_bin, open(prefix + '_domains.bed', 'w') as file_domains, open(prefix + '_boundaries.gff', 'w') as gff:
        for idx, min_bin_id in enumerate(filtered_min_idx):
            # skip if the start of the boundary
            # is the end of the chromosome
            if min_bin_id in chr_end_idx:
                continue

            if min_bin_id not in delta_of_min:
                delta_of_min[min_bin_id] = np.nan

            # 1. save boundaries at 1bp position
            right_bin_center = chr_start[min_bin_id] + int((chr_end[min_bin_id] - chr_start[min_bin_id]) / 2)

            left_bin_center = chr_start[min_bin_id - 1] + int((chr_end[min_bin_id - 1] - chr_start[min_bin_id - 1]) / 2)

            if chrom[min_bin_id] != chrom[min_bin_id - 1]:
                continue

            # 2. save the position of the boundary range
            file_boundary_bin.write("{}\t{}\t{}\tB{:05d}\t{}\t.\n".format(chrom[min_bin_id],
                                                                          left_bin_center,
                                                                          right_bin_center,
                                                                          min_bin_id,
                                                                          mean_mat_all[min_bin_id]))

            # safe gff file that can contain more information
            gff.write("{chrom}\tHiCExplorer\tboundary\t{start}\t{end}\t{score}"
                      "\t.\t.\tID=B{id:05d};delta={delta};pvalue={pvalue};tad_sep={score}\n".format(chrom=chrom[min_bin_id],
                                                                                                    start=left_bin_center,
                                                                                                    end=right_bin_center,
                                                                                                    delta=delta_of_min[min_bin_id],
                                                                                                    pvalue=pvalue_of_min[min_bin_id],
                                                                                                    score=mean_mat_all[min_bin_id],
                                                                                                    id=min_bin_id))

            start = chr_start[min_bin_id]
            # check that the next boundary exists and is in the same chromosome
            if idx + 1 == len(filtered_min_idx) or chrom[min_bin_id] != chrom[filtered_min_idx[idx + 1]]:
                continue

            end = chr_start[filtered_min_idx[idx + 1]]

            # 2. save domain intervals
            if count % 2 == 0:
                rgb = '51,160,44'
            else:
                rgb = '31,120,180'

            file_domains.write("{0}\t{1}\t{2}\tID_{6}_{3}\t{4}\t.\t{1}\t{2}\t{5}\n".format(chrom[min_bin_id],
                                                                                           start, end, count,
                                                                                           mean_mat_all[min_bin_id],
                                                                                           rgb, args.delta))

            count += 1

    # save track with mean values in bedgraph format
    with open(args.outPrefix + '_score.bedgraph', 'w') as tad_score:
        for idx in range(1, len(chrom)):
            right_bin_center = chr_start[idx] + int((chr_end[idx] - chr_start[idx]) / 2)
            left_bin_center = chr_start[idx - 1] + int((chr_end[idx - 1] - chr_start[idx - 1]) / 2)
            if right_bin_center < left_bin_center:
                # this condition happens at chromosome borders
                continue
            tad_score.write("{}\t{}\t{}\t{}\n".format(chrom[idx], left_bin_center, right_bin_center,
                                                      mean_mat_all[idx]))


def compute_spectra_matrix(args, matrix=None):

    if args.maxDepth is not None and args.minDepth is not None and args.maxDepth <= args.minDepth:
        exit("Please check that maxDepth is larger than minDepth.")

    global hic_ma
    if matrix is not None:
        hic_ma = matrix
    else:
        hic_ma = hm.hiCMatrix(args.matrix)

    # remove self counts
    log.info('removing diagonal values\n')
    hic_ma.diagflat(value=0)

    # mask bins without any information
    hic_ma.maskBins(hic_ma.nan_bins)
    orig_intervals = hic_ma.cut_intervals

    binsize = hic_ma.getBinSize()

    if args.maxDepth is None:
        if binsize < 1000:
            args.maxDepth = binsize * 60
        elif 1000 <= binsize < 20000:
            args.maxDepth = binsize * 20
        else:
            args.maxDepth = binsize * 10
    elif args.maxDepth < binsize * 5:
        sys.error("Please specify a --maxDepth that is at least 5 times larger than the matrix bin size")
        exit(1)

    if args.minDepth is None:
        if binsize < 1000:
            args.minDepth = binsize * 30
        elif 1000 <= binsize < 20000:
            args.minDepth = binsize * 10
        else:
            args.minDepth = binsize * 5
    elif args.minDepth < binsize * 3:
        log.error("Please specify a --minDepth that is at least 3 times larger than the matrix bin size")
        exit(1)

    if args.step is None:
        if binsize < 1000:
            args.step = binsize * 4
        else:
            args.step = binsize * 2

    elif args.step < binsize:
        log.error("Please specify a --step that is at least the size of the matrix bin size")
        exit(1)

    args.binsize = binsize
    print_args(args)

    # use zscore matrix
    log.info("Computing z-score matrix...\n")
    hic_ma.convert_to_zscore_matrix(maxdepth=args.maxDepth * 2.5, perchr=True)

    # extend remaining bins to remove gaps in
    # the matrix
    new_intervals = enlarge_bins(hic_ma.cut_intervals)

    # rebuilt bin positions if necessary

    if new_intervals != orig_intervals:
        hic_ma.interval_trees, hic_ma.chrBinBoundaries = \
            hic_ma.intervalListToIntervalTree(new_intervals)
        hic_ma.cut_intervals = new_intervals
        hic_ma.orig_bin_ids = None
        hic_ma.nan_bins = None

    hic_ma.save(args.outFileName + "_zscore_matrix.h5")

    if args.minDepth % hic_ma.getBinSize() != 0:
        log.warn('Warning. specified *depth* is not multiple of the '
                 'hi-c matrix bin size ({})\n'.format(hic_ma.getBinSize()))
    if args.step % hic_ma.getBinSize() != 0:
        log.warn('Warning. specified *step* is not multiple of the '
                 'hi-c matrix bin size ({})\n'.format(hic_ma.getBinSize()))

    binsize = hic_ma.getBinSize()

    log.info("Computing TAD-separation scores...\n")
    min_depth_in_bins = int(args.minDepth / binsize)
    max_depth_in_bins = int(args.maxDepth / binsize)
    step_in_bins = int(args.step / binsize)
    if step_in_bins == 0:
        exit("Please select a step size larger than {}".format(binsize))

    incremental_step = get_incremental_step_size(args.minDepth, args.maxDepth, args.step)

    log.info("computing spectrum for window sizes between {} ({} bp)"
             "and {} ({} bp) at the following window sizes {} {}\n".format(min_depth_in_bins,
                                                                           binsize * min_depth_in_bins,
                                                                           max_depth_in_bins,
                                                                           binsize * max_depth_in_bins,
                                                                           step_in_bins, incremental_step))
    if min_depth_in_bins <= 1:
        log.error('ERROR\nminDepth length too small. Use a value that is at least '
                  'twice as large as the bin size which is: {}\n'.format(binsize))
        exit(0)

    if max_depth_in_bins <= 1:
        log.error('ERROR\nmaxDepth length too small. Use a value that is larger '
                  'than the bin size which is: {}\n'.format(binsize))
        exit(0)

    # work only with the upper matrix
    # and remove all pixels that are beyond
    # 2 * max_depth_in_bis which are not required
    # (this is done by subtracting a second sparse matrix
    # that contains only the upper matrix that wants to be removed.
    limit = 2 * max_depth_in_bins
    hic_ma.matrix = sparse.triu(hic_ma.matrix, k=0, format='csr') - sparse.triu(hic_ma.matrix, k=limit, format='csr')
    hic_ma.matrix.eliminate_zeros()

    num_processors = args.numberOfProcessors

    func = compute_matrix_wrapper
    TASKS = []
    bins_to_consider = []
    for chrom in hic_ma.chrBinBoundaries.keys():
        bins_to_consider.extend(range(*hic_ma.chrBinBoundaries[chrom]))

    for idx_array in np.array_split(bins_to_consider, num_processors):
        TASKS.append((idx_array, args.minDepth, args.maxDepth, args.step))

    if num_processors > 1:
        pool = multiprocessing.Pool(num_processors)
        log.info("Using {} processors\n".format(num_processors))
        res = pool.map_async(func, TASKS).get(9999999)
    else:
        res = map(func, TASKS)

    chrom = []
    chr_start = []
    chr_end = []
    matrix = []
    for _chrom, _chr_start, _chr_end, _matrix in res:
        chrom.extend(_chrom)
        chr_start.extend(_chr_start)
        chr_end.extend(_chr_end)
        matrix.append(_matrix)

    matrix = np.vstack(matrix)
    return np.array(chrom), np.array(chr_start).astype(int), np.array(chr_end).astype(int), matrix


def load_spectrum_matrix(file):
    # load spectrum matrix:
    matrix = []
    chrom_list = []
    start_list = []
    end_list = []
    with open(file, 'r') as fh:
        for line in fh:
            if line.startswith("#"):
                # recover the parameters used to generate the spectrum_matrix
                parameters = json.loads(line[1:].strip())
                continue
            fields = line.strip().split('\t')
            chrom, start, end = fields[0:3]
            chrom_list.append(chrom)
            start_list.append(int(float(start)))
            end_list.append(int(float(end)))
            matrix.append(map(float, fields[3:]))

    matrix = np.vstack(matrix)
    chrom = np.array(chrom_list)
    start = np.array(start_list)
    end = np.array(end_list)
    return chrom, start, end, matrix, parameters


def print_args(args):
    """
    Print to stderr the parameters used
    Parameters
    ----------
    args

    Returns
    -------

    """
    for key, value in args._get_kwargs():
        sys.stderr.write("{}:\t{}\n".format(key, value))


def min_pvalue(matrix_file, min_idx, chrom, chr_start, chr_end, window_len=20000, delta=None):
    """
    For each putative local minima, find the -window_len diammond and the +window_len diamond
    and compare with the local minima using wilcoxon rank sum.

    Parameters
    ----------
    min_idx list of local minima
    window_len in base pairs

    Returns
    -------
    list of p-values per each local minima

    """
    hic_ma = hm.hiCMatrix(matrix_file)

    log.info("Computing p-values for window length: {}\n".format(window_len))
    pvalues = []

    from scipy.stats import ranksums

    new_min_idx = []
    for idx in min_idx:

        matrix_idx = hic_ma.getRegionBinRange(chrom[idx], chr_start[idx], chr_end[idx])
        if matrix_idx is None:
            continue
        else:
            matrix_idx = matrix_idx[0]

        new_min_idx += [idx]
        min_chr, min_start, min_end, _ = hic_ma.getBinPos(matrix_idx)
        assert chrom[idx] == min_chr and chr_start[idx] == min_start and chr_end[idx] == min_end

        left_idx, right_idx = get_idx_of_bins_at_given_distance(hic_ma, matrix_idx, window_len)

        left = get_cut_weight(hic_ma, left_idx, window_len)
        right = get_cut_weight(hic_ma, right_idx, window_len)
        boundary = get_cut_weight(hic_ma, matrix_idx, window_len)
        if left is None:
            left = []
        if right is None:
            right = []

        if len(left) == 0 and len(right) == 0:
            pval = np.nan

        elif boundary is None or len(boundary) == 0:
            pval = np.nan
        else:
            try:
                pval1 = ranksums(boundary, left)[1]
                pval2 = ranksums(boundary, right)[1]
                pval = min(pval1, pval2)
            except ValueError:
                # this condition happens when boundary, left and right are only composed of 'zeros'
                pval = np.nan

        pvalues += [pval]
        # if delta[idx] > 0.04 and pval > 0.001:
        #    import ipdb; ipdb.set_trace()

    assert len(pvalues) == len(new_min_idx)
    # bonferroni correction
    pvalues = np.array(pvalues) * len(pvalues)
    pvalues[pvalues > 1] = 1
    return dict(zip(new_min_idx, pvalues))


def min_pvalue_diamond_vs_triangle(matrix_file, min_idx, chrom, chr_start, chr_end, window_len=20000, delta=None):
    """
    For each putative local minima, find the -window_len triangle and the +window_len triangle
    and compare with the local minima dianmond using wilcoxon rank sum.

    Parameters
    ----------
    min_idx list of local minima
    window_len in base pairs

    Returns
    -------
    list of p-values per each local minima

    """
    hic_ma = hm.hiCMatrix(matrix_file)

    log.info("Computing p-values for window length: {}\n".format(window_len))
    pvalues = []

    from scipy.stats import ranksums

    new_min_idx = []
    for idx in min_idx:

        matrix_idx = hic_ma.getRegionBinRange(chrom[idx], chr_start[idx], chr_end[idx])
        if matrix_idx is None:
            continue
        else:
            matrix_idx = matrix_idx[0]

        new_min_idx += [idx]
        min_chr, min_start, min_end, _ = hic_ma.getBinPos(matrix_idx)
        assert chrom[idx] == min_chr and chr_start[idx] == min_start and chr_end[idx] == min_end

        triangle = get_triangle(hic_ma, matrix_idx, window_len)
        boundary = get_cut_weight(hic_ma, matrix_idx, window_len)
        if triangle is None or len(triangle) == 0:
            pval = np.nan
        elif boundary is None or len(boundary) == 0:
            pval = np.nan
        else:
            try:
                pval = ranksums(boundary[boundary.nonzero()], triangle[triangle.nonzero()])[1]
            except ValueError:
                # this condition happens when boundary, left and right are only composed of 'zeros'
                pval = np.nan

        pvalues += [pval]

    assert len(pvalues) == len(new_min_idx)
    # bonferroni correction
    pvalues = np.array(pvalues) * len(pvalues)
    pvalues[pvalues > 1] = 1

    return dict(zip(new_min_idx, pvalues))


def main(args=None):

    args = parse_arguments().parse_args(args)
    if args.command == 'TAD_score':
        chrom, chr_start, chr_end, matrix = compute_spectra_matrix(args)
        save_bedgraph_matrix(args.outFileName, chrom, chr_start, chr_end, matrix, args)
        return

    chrom, chr_start, chr_end, matrix, parameters = load_spectrum_matrix(args.tadScoreFile)

    # perform some checks
    avg_bin_size = np.median(chr_end - chr_start)

    # compute lookahead (in number of bins)
    if args.minBoundaryDistance is None:
        args.minBoundaryDistance = avg_bin_size * 4
    print_args(args)

    lookahead = int(args.minBoundaryDistance / avg_bin_size)
    if lookahead < 1:
        raise (ValueError, "minBoundaryDistance must be '1' or above in value")

    min_idx, delta = find_consensus_minima(matrix, lookahead=lookahead, chrom=chrom)

    pvalues = min_pvalue(parameters['zscore_matrix'], min_idx, chrom, chr_start, chr_end,
                         window_len=parameters['minDepth'], delta=delta)

    # pvalues = min_pvalue_diamond_vs_triangle(parameters['zscore_matrix'], min_idx, chrom, chr_start, chr_end,
    #                     window_len=parameters['minDepth'], delta=delta)
    if len(min_idx) <= 10:
        mat_mean = matrix.mean(axis=1)
        m_mean = mat_mean.mean()
        m_median = np.median(mat_mean)
        m_75 = np.percentile(mat_mean, 75)
        m_25 = np.percentile(mat_mean, 25)

        msg = ("Please check the parameters:\n"
               " delta: {}\n"
               " minBoundaryDistance: {}\n\n"
               "TAD Score values:\n"
               " mean: {:.3f}\n"
               " median: {:.3f}\n"
               " 1st quartile: {:.3f}\n"
               " 3rd quartile: {:.3f}\n".format(args.delta, args.minBoundaryDistance, m_mean, m_median, m_25, m_75))

        if len(min_idx) == 0:
            exit("\n*ERROR*\nNo boundaries were found. {}".format(msg))
        else:
            log.info("Only {} boundaries found. {}".format(len(min_idx), msg))

    save_domains_and_boundaries(chrom, chr_start, chr_end, matrix, min_idx, delta, pvalues, args)

    # turn of hierarchical clustering which is apparently not working.

    # if 2 == 1:
    #     boundary_list = [(hic_ma.cut_intervals[min_][0], hic_ma.cut_intervals[min_][2], mean_mat[min_]) for min_ in min_idx]

    #     Z, clusters = hierarchical_clustering(boundary_list, clusters_cutoff=[0.4, 0.3, 0.2])

    #     save_linkage(Z, args.outPrefix + '_linkage.bed')
    #     save_clusters(clusters, args.outPrefix)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import os
import errno
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from hicexplorer._version import __version__


def parse_arguments():
    parser = argparse.ArgumentParser(description='Tabulates and plots QC measures from  '
                                                 'hicBuildMatrix log files.',
                                     usage='%(prog)s --logfiles matrix1.h5 matrix2.h5 '
                                           '--labels "sample 1" "sample 2" -outFile QC.txt)')

    # define the arguments
    parser.add_argument('--logfiles', '-l',
                        help='Path to the log files to be processed',
                        type=argparse.FileType('r'),
                        nargs="+",
                        required=True)

    parser.add_argument('--labels',
                        help='Label to assign to each log file. Each label should be separated by a space. Quote '
                             'labels that contain spaces: E.g. --labels label1 "labels 2"',
                        nargs="+")

    parser.add_argument('--outputFolder', '-o',
                        help='Several files with be saved under this folder: A table containing the results and '
                             'a html file with several images.',
                        required=True)

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def save_html(filename, html_table):
    root = os.path.dirname(os.path.abspath(__file__))

    html = open(os.path.join(root, "qc_template.html"), "r").read()
    # the html code has a placeholder for the html table
    html = html.replace("%%TABLE%%", html_table)
    with open(filename, 'w') as fh:
        fh.write(html)


def make_sure_path_exists(path):
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def make_figure_pairs_considered(table, filename):

    fig = plt.figure(figsize=(8, 3))
    ax = fig.add_subplot(111)
    table['Pairs considered'].plot.barh(ax=ax, color='#999999')
    xticks = ax.get_xticks()
    if xticks[-1] > 1e6:
        labels = ["{:.0f}".format(float(x) / 1e6) for x in xticks]
        labels[-1] += 'M'
    else:
        labels = ["{:,}".format(int(x)) for x in xticks]

    ax.set_xticklabels(labels)
    ax.set_ylabel("")
    ax.set_xlabel("Number of reads")
    plt.tight_layout()
    plt.savefig(filename)


def make_figure_pairs_used(table, filename):
    prc_table = table[['Pairs used', 'One mate low quality', 'One mate not unique',
                       'One mate not close to rest site', 'One mate unmapped', 'dangling end', 'duplicated pairs',
                       'same fragment (800 bp)', 'self circle',
                       'self ligation (removed)']].T / table['Pairs considered']

    fig = plt.figure(figsize=(7, 5))
    ax = fig.add_subplot(111)
    prc_table.plot.bar(ax=ax)
    labels = ax.get_xticklabels()
    ax.set_xticklabels(labels, rotation=45, ha='right')
    handles, labels = ax.get_legend_handles_labels()
    lgd = ax.legend(handles, labels, loc='center left', bbox_to_anchor=(1, 0.5))
    ax.set_ylabel("fraction")
    plt.savefig(filename, bbox_extra_artists=(lgd,), bbox_inches='tight')


def make_figure_distance(table, filename):

    prc_table2 = table[['inter chromosomal', 'short range < 20kb', 'long range']].T / table['Pairs used']
    fig = plt.figure(figsize=(4, 5))
    ax = fig.add_subplot(111)
    prc_table2.plot.bar(ax=ax)
    labels = ax.get_xticklabels()
    ax.set_xticklabels(labels, rotation=45, ha='right')
    handles, labels = ax.get_legend_handles_labels()
    lgd = ax.legend(handles, labels, loc='center left', bbox_to_anchor=(1, 0.5))
    ax.set_ylabel("fraction")

    plt.savefig(filename, bbox_extra_artists=(lgd,), bbox_inches='tight')


def make_figure_read_orientation(table, filename):
    _t = table[[u'inward pairs', u'outward pairs', u'left pairs', u'right pairs']].T
    prc_table3 = _t / _t.sum(axis=0)
    fig = plt.figure(figsize=(4, 3))
    ax = fig.add_subplot(111)
    prc_table3.plot.bar(ax=ax)
    handles, labels = ax.get_legend_handles_labels()
    lgd = ax.legend(handles, labels, loc='center left', bbox_to_anchor=(1, 0.5))
    ax.set_ylabel("fraction")
    plt.savefig(filename, bbox_extra_artists=(lgd,), bbox_inches='tight')


def main(args=None):

    """
    The structure of the log file is as follows:
    --------------------------------------------

    File    hic/Beaf32_KD_repA_rf.h5
    Pairs considered        102449560
    Min rest. site distance 150
    Max rest. site distance 1500


    Pairs used      14537174        (14.19) (14.62)
    One mate unmapped       3031462 (2.96)  (3.05)
    One mate not unique     29915787        (29.20) (30.09)
    One mate low quality    6471448 (6.32)  (6.51)
    dangling end    2398169 (2.34)  (2.41)
    self ligation (removed) 14263507        (13.92) (14.35)
    One mate not close to rest site 371977  (0.36)  (0.37)
    same fragment (800 bp)  28543891        (27.86) (28.71)
    self circle     2845649 (2.78)  (2.86)
    duplicated pairs        2916144 (2.85)  (2.93)
    Of pairs used:
    inter chromosomal       951976  (6.55)
    short range < 20kb      9337137 (64.23)
    long range      4248061 (29.22)
    inward pairs    2686016 (18.48)
    outward pairs   3709843 (25.52)
    left pairs      3589334 (24.69)
    right pairs     3600005 (24.76)

    """

    args = parse_arguments().parse_args(args)
    params = dict()
    make_sure_path_exists(args.outputFolder)
    for fh in args.logfiles:
        in_log_part = False
        for line in fh.readlines():
            if line.startswith("File"):
                in_log_part = True
            if in_log_part is True:
                if line.strip() == "":
                    continue
                fields = line.strip().split("\t")
                if len(fields) == 1:
                    continue
                if fields[0] not in params:
                    params[fields[0]] = []

                try:
                    params[fields[0]].append(int(fields[1]))
                except ValueError:
                    params[fields[0]].append(fields[1])

    import pandas as pd
    table = pd.DataFrame(params)
    if args.labels and len(args.labels) == len(args.logfiles):
            table['Labels'] = args.labels
            table = table.set_index('Labels')
    else:
        table = table.set_index('File')

    table.to_csv(args.outputFolder + "/table.txt", sep="\t")

    make_figure_pairs_considered(table, args.outputFolder + "/pairs_considered.png")
    make_figure_pairs_used(table, args.outputFolder + "/pairs_used.png")
    make_figure_distance(table, args.outputFolder + "/distance.png")
    make_figure_read_orientation(table, args.outputFolder + "/read_orientation.png")

    save_html(args.outputFolder + "/hicQC.html", table.T.to_html())



import sys
import argparse
import re
from tempfile import NamedTemporaryFile
import subprocess

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.Alphabet import generic_dna


def parse_arguments(args=None):
    parser = argparse.ArgumentParser(description='Identifies the genomic locations of restriction sites. ',
                                     usage='%(prog)s --fasta mm10.fa '
                                           '--searchPattern AAGCTT -o rest_site_positions.bed')

    # define the arguments
    parser.add_argument('--fasta', '-f',
                        help='Path to fasta file for the organism genome.',
                        type=argparse.FileType('r'),
                        required=True)

    # define the arguments
    parser.add_argument('--searchPattern', '-p',
                        help='Search pattern. For example, for HindII this pattern is "AAGCTT". '
                             'Both, forward and reverse strand are searched for a match. The pattern '
                             'is a regexp and can contain regexp specif syntax '
                             '(see https://docs.python.org/2/library/re.html). For example the pattern'
                             'CG..GC will find all occurrence of CG followed by any two bases and then GC.',
                        required=True)

    parser.add_argument('--outFile', '-o',
                        help='Name for the resulting bed file.',
                        type=argparse.FileType('w'),
                        required=True)

    return parser


def find_pattern(pattern, fasta_file, out_file):
    r"""
    Finds the occurrences of the match in the fasta file
    and saves a bed file.

    The coordinate system is zero based

    :param pattern: Sequence to search for
    :param fasta_file:
    :param out_file: file handler

    :return: none

    This is an example from Emboss remap for Csp6I with
    restriction site GTAC


                       Csp6I
               Csp6I   |    Csp6I Csp6I
               \       \    \     \
          CTACGGTACGAACGTACGGTACGcgtaCGNAGTCATG
                   10        20        30
          ----:----|----:----|----:----|----:--
          GATGCCATGCTTGCATGCCATGCgcatGCNTCAGTAC
                 /       /    /     /
                 Csp6I   |    Csp6I Csp6I
                         Csp6I

    >>> fa = open("/tmp/test.fa", 'w')
    >>> fa.write(">chr1\nCTACGGTACGAACGTACGGTACGcgtaCGNAGTCATG\n")
    >>> fa.close()
    >>> find_pattern("GTAC", "/tmp/test.fa", open("/tmp/test.bed", 'w'))
    >>> open("/tmp/test.bed", 'r').readlines()
    ['chr1\t5\t9\t.\t0\t+\n', 'chr1\t13\t17\t.\t0\t+\n', 'chr1\t18\t22\t.\t0\t+\n', 'chr1\t24\t28\t.\t0\t+\n']

    Test with non palindromic sequence with regexp
    >>> find_pattern("CG.AG", "/tmp/test.fa", open("/tmp/test.bed", 'w'))
    >>> open("/tmp/test.bed", 'r').readlines()
    ['chr1\t0\t5\t.\t0\t-\n', 'chr1\t27\t32\t.\t0\t+\n']

    """

    # get the reverse complement of the pattern
    rev_compl = str(Seq(pattern, generic_dna).reverse_complement())

    temp = NamedTemporaryFile(suffix=".bed", delete=False)
    for record in SeqIO.parse(fasta_file, 'fasta', generic_dna):
        # find all the occurrences of pattern
        for match in re.finditer(pattern, str(record.seq), re.IGNORECASE):
            temp.write('{}\t{}\t{}\t.\t0\t+\n'.format(record.name,
                                                      match.start(),
                                                      match.end()))
        if rev_compl != pattern:
            # search for the reverse complement only if the pattern is not palindromic
            for match in re.finditer(rev_compl, str(record.seq), re.IGNORECASE):
                temp.write('{}\t{}\t{}\t.\t0\t-\n'.format(record.name,
                                                          match.start(),
                                                          match.end()))

    sys.stderr.write("Sorting file ...\n")
    tmpfile_name = temp.name
    temp.close()
    subprocess.check_output(["cat", tmpfile_name])
    # sort bed file using system tools
    cmd = 'sort -k1,1 -k2,2n -u {}'.format(tmpfile_name)
    # LC_ALL=C is to set the appropriate collation order
    proc = subprocess.Popen(cmd.split(" "), stdout=subprocess.PIPE, env={'LC_ALL': ' C'})
    stdout, _ = proc.communicate()

    out_file.write(stdout)
    out_file.close()

    temp.close()


def main():
    args = parse_arguments().parse_args()
    find_pattern(args.searchPattern, args.fasta, args.outFile)



import argparse
import numpy as np
from hicexplorer import HiCMatrix as hm
from hicexplorer._version import __version__


def parse_arguments(args=None):

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=('Takes two matrices, normalizes them and applies'
                                                  'the given operation. To normalize the matrices '
                                                  'each element is divided by sum of the matrix.'))

    parser.add_argument('--matrices', '-m',
                        help='matrices to use.',
                        metavar='.h5 file format',
                        nargs=2,
                        required=True)

    parser.add_argument('--outFileName', '-o',
                        help='File name to save the resulting matrix. The output is '
                             'also a .h5 file.',
                        required=True)

    parser.add_argument('--operation',
                        help='Operation to apply for the matrices. Options are: diff, ratio, log2ratio',
                        required=True)

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def main():

    args = parse_arguments().parse_args()
    hic1 = hm.hiCMatrix(args.matrices[0])
    hic2 = hm.hiCMatrix(args.matrices[1])

    if hic1.matrix.shape != hic2.matrix.shape:
        exit("The two matrices have different size. Use matrices having the same resolution and created using"
             "the same parameters. Check the matrix values using the tool `hicInfo`.")

    if hic1.chrBinBoundaries != hic2.chrBinBoundaries:
        exit("The two matrices have different chromosome order. Use the tool `hicExport` to change the order.\n"
             "{}: {}\n"
             "{}: {}".format(args.matrices[0], hic1.chrBinBoundaries.keys(),
                             args.matrices[1], hic2.chrBinBoundaries.keys()))

    # normalize by total matrix sum
    hic1.matrix.data = hic1.matrix.data.astype(float) / hic1.matrix.data.sum()
    hic2.matrix.data = hic2.matrix.data.astype(float) / hic2.matrix.data.sum()

    nan_bins = set(hic1.nan_bins)
    nan_bins = nan_bins.union(hic2.nan_bins)

    if args.operation == 'diff':
        new_matrix = hic1.matrix - hic2.matrix
    else:
        hic2.matrix.data = float(1) / hic2.matrix.data
        new_matrix = hic1.matrix.multiply(hic2.matrix)
        # just in case
        new_matrix.eliminate_zeros()
        if args.operation == 'log2ratio':
            new_matrix.data = np.log2(new_matrix.data)
            new_matrix.eliminate_zeros()

    hic1.setMatrixValues(new_matrix)
    hic1.maskBins(sorted(nan_bins))
    hic1.save(args.outFileName)



# -*- coding: utf-8 -*-

import os
import sys
import subprocess
import re

from setuptools import setup
from setuptools.command.sdist import sdist as _sdist
from setuptools.command.install import install as _install

VERSION_PY = """
# This file is originally generated from Git information by running 'setup.py
# version'. Distribution tarballs contain a pre-generated copy of this file.

__version__ = '%s'
"""


def update_version_py():
    if not os.path.isdir(".git"):
        print("This does not appear to be a Git repository.")
        return
    try:
        p = subprocess.Popen(["git", "describe",
                              "--tags", "--always"],
                             stdout=subprocess.PIPE)
    except EnvironmentError:
        print("unable to run git, leaving hicexplorer/_version.py alone")
        return
    stdout = p.communicate()[0]
    if p.returncode != 0:
        print("unable to run git, leaving hicexplorer/_version.py alone")
        return
    ver = stdout.strip()
    f = open("hicexplorer/_version.py", "w")
    f.write(VERSION_PY % ver)
    f.close()
    print("set hicexplorer/_version.py to '%s'" % ver)


def get_version():
    try:
        f = open("hicexplorer/_version.py")
    except EnvironmentError:
        return None
    for line in f.readlines():
        mo = re.match("__version__ = '([^']+)'", line)
        if mo:
            ver = mo.group(1)
            return ver
    return None


class sdist(_sdist):

    def run(self):
        update_version_py()
        self.distribution.metadata.version = get_version()
        return _sdist.run(self)

# Install class to check for external dependencies from OS environment


class install(_install):

    def run(self):
        update_version_py()
        self.distribution.metadata.version = get_version()
        _install.run(self)
        return

    def checkProgramIsInstalled(self, program, args, where_to_download,
                                affected_tools):
        try:
            subprocess.Popen([program, args],
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE)
            return True
        except EnvironmentError:
            # handle file not found error.
            # the config file is installed in:
            msg = "\n**{0} not found. This " \
                  "program is needed for the following "\
                  "tools to work properly:\n"\
                  " {1}\n"\
                  "{0} can be downloaded from here:\n " \
                  " {2}\n".format(program, affected_tools,
                                  where_to_download)
            sys.stderr.write(msg)

        except Exception as e:
            sys.stderr.write("Error: {}".format(e))


setup(
    name='HiCExplorer',
    version=get_version(),
    author='Fidel Ramirez, Vivek Bhardwaj, Björn Grüning',
    author_email='deeptools@googlegroups.com',
    packages=['hicexplorer'],
    scripts=['bin/findRestSite', 'bin/hicBuildMatrix', 'bin/hicCorrectMatrix',
             'bin/hicCorrelate', 'bin/hicFindEnrichedContacts', 'bin/hicFindTADs',
             'bin/hicMergeMatrixBins', 'bin/hicPlotMatrix', 'bin/hicPlotDistVsCounts',
             'bin/hicPlotTADs', 'bin/hicSumMatrices', 'bin/hicExport', 'bin/hicInfo', 'bin/hicexplorer',
             'bin/hicQC', 'bin/hicCompareMatrices'],
    include_package_data=True,
    package_dir={'hicexplorer': 'hicexplorer'},
    package_data={'hicexplorer': ['qc_template.html']},
    url='http://hicexplorer.readthedocs.io',
    license='LICENSE.txt',
    description='Set of programs to process, analyze and visualize Hi-C data',
    long_description=open('README.rst').read(),
    install_requires=[
        "numpy >= 1.10.4",
        "scipy >= 0.17.1",
        "matplotlib >= 1.5.3",
        "pysam >= 0.8.3",
        "intervaltree >= 2.1.0",
        "biopython >= 1.65",
        "tables >= 3.2.2",
        "pandas >= 0.19.2",
        "pyBigWig >=0.2.8"],
    zip_safe=False,
    cmdclass={'sdist': sdist, 'install': install}
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import argparse
import numpy as np
from hicexplorer import HiCMatrix as Hm


def parse_arguments():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
Uses a BED file of domains or TAD boundaries to merge
the bin counts of a HiCMatrix.

In this matrix only contains the total counts in each TAD and
its total contacts of all other TADs.""")

    # define the arguments
    parser.add_argument('--matrix', '-m',
                        help='Path to Hi-C matrix to use',
                        required=True)

    parser.add_argument('--domains',
                        help='Path to a bed file containing the domains.',
                        type=argparse.FileType('r'),
                        required=True)

    parser.add_argument('--outFile', '-o',
                        help='Name for the resulting matrix file.',
                        required=True)

    return parser


def merge_tad_bins(hic, boundary_id_list, filename):
    """
    Reduces the HiC matrix by merging the counts of tad bins.
    :param hic: hicMatrix object
    :param boundary_id_list list of tad boundary bin ids
    :param filename Name to save the resulting matrix
    :return: hicMatrix object
    """

    from hicexplorer.reduceMatrix import reduce_matrix
    hic.restoreMaskedBins()
    ref_name_list, start_list, end_list, coverage_list = zip(*hic.cut_intervals)
    new_bins = []
    bins_to_merge = []
    prev_ref = ref_name_list[0]

    # prepare new intervals
    idx_start = 0
    new_start = start_list[0]
    count = 0
    for idx, ref in enumerate(ref_name_list):
        if (count > 0 and idx in boundary_id_list) or ref != prev_ref:
            coverage = np.mean(coverage_list[idx_start:idx])
            new_bins.append((ref_name_list[idx_start], new_start,
                             end_list[idx - 1], coverage))
            bins_to_merge.append(range(idx_start, idx))
            idx_start = idx
            new_start = start_list[idx]
            count = 0

        prev_ref = ref
        count += 1
    # check that the previous for loop ran, otherwise
    # some variables may not be set
    if len(bins_to_merge) > 0:
        coverage = np.mean(coverage_list[idx_start:])
        new_bins.append((ref, new_start, end_list[idx], coverage))
        bins_to_merge.append(range(idx_start, idx + 1))
        # remove correction factors otherwise they are
        # saved but they no longer correspond to the
        # size of the matrix.
        hic.correction_factors = None

        hic.update_matrix(
            reduce_matrix(hic.matrix, bins_to_merge, diagonal=True), new_bins)

        hic.save(filename)
    else:
        sys.stderr.write("Nothing to merge.")


def get_boundary_bin_id(hic, bed_fh):
    """
    :param hic: HiCMatrix object
    :param bed_fh: file handle of the bed file
    :return: Sorted list of bin indices.
    """
    line_number = 0
    boundaries = set()
    for line in bed_fh.readlines():
        line_number += 1
        if line.startswith('browser') or line.startswith('track') or line.startswith('#'):
            continue
        try:
            chrom, start, end = line.strip().split('\t')[0:3]
        except Exception as detail:
            msg = 'Could not read line\n{}\n. {}'.format(line, detail)
            sys.exit(msg)

        try:
            start = int(start)
            end = int(end)
        except ValueError as detail:
            msg = "Error reading line: {}. One of the fields is not " \
                  "an integer.\nError message: {}".format(line_number, detail)
            sys.exit(msg)

        assert start <= end, "Error in line #{}, end1 larger than start1 in {}".format(line_number, line)

        # check the overlap of the region with the hic matrix bins
        start_bin, end_bin = hic.getRegionBinRange(chrom, start, end)
        boundaries.add(start_bin)
        boundaries.add(end_bin)

    return np.sort(list(boundaries))


def main(args=None):

    args = parse_arguments().parse_args(args)
    hic_ma = Hm.hiCMatrix(args.matrix)
    hic_ma.restoreMaskedBins()

    # the bin id of boundary positions
    boundary_id_list = get_boundary_bin_id(hic_ma, args.domains)

    # make a reduce matrix by merging the TAD bins
    sys.stderr.write("Generating matrix with merged bins\n")
    merge_tad_bins(hic_ma, boundary_id_list, args.outFile)
